"""
A little utiltity for performing C-like preprocessing using standard CPP
directives like #if, #ifdef, and #define.

Written by Jason Swails
"""
from parmed.exceptions import PreProcessorError, PreProcessorWarning
from parmed.utils.io import genopen
from parmed.utils.six import string_types, iteritems, wraps
from collections import OrderedDict
from os import path
import re
import warnings

ppre = re.compile(r'#\s*(ifdef|ifndef|if|else|elif|endif|define|undef|include)'
                  r'\s*(.+)?')
ppcomments = re.compile(r'(?://.+|/\*(?:.*)\*/)')
includere = re.compile(r'[<"](.+)[>"]')
novarcharre = re.compile(r'\W')

def _strip_pp_comments(func):
    """
    Decorator to apply to functions that will strip out C-style comments before
    calling the preprocessor function on the resulting arguments to the
    preprocessor directive
    """
    @wraps(func)
    def wrapper(self, args):
        args = ppcomments.sub('', args)
        return func(self, args)
    return wrapper

def _find_all_instances_in_string(string, substr):
    """ Find indices of all instances of substr in string """
    indices = []
    idx = string.find(substr, 0)
    while idx > -1:
        indices.append(idx)
        idx = string.find(substr, idx+1)
    return indices

def _replace_defines(line, defines):
    """ Replaces defined tokens in a given line """
    if not defines: return line
    for define in reversed(defines):
        value = defines[define]
        indices = _find_all_instances_in_string(line, define)
        if not indices: continue
        # Check to see if it's inside of quotes
        inside = ''
        idx = 0
        n_to_skip = 0
        new_line = []
        for i, char in enumerate(line):
            if n_to_skip:
                n_to_skip -= 1
                continue
            if char in ('\'"'):
                if not inside:
                    inside = char
                else:
                    if inside == char:
                        inside = ''
            if idx < len(indices) and i == indices[idx]:
                if inside:
                    new_line.append(char)
                    idx += 1
                    continue
                if i == 0 or novarcharre.match(line[i-1]):
                    endidx = indices[idx] + len(define)
                    if endidx >= len(line) or novarcharre.match(line[endidx]):
                        new_line.extend(list(value))
                        n_to_skip = len(define) - 1
                        idx += 1
                        continue
                idx += 1
            new_line.append(char)
        line = ''.join(new_line)
                        
    return line

# To track where in the "if-elif-else" block each conditional is
_IN_IF = 'in if'
_IN_ELIF = 'in elif'
_IN_ELSE = 'in else'

class CPreProcessor(object):
    """
    Steps through a file line-by-line, yielding only the preprocessed contents

    Parameters
    ----------
    fname : str or file-like
        The file name of the file to open or the open file object
    defines : dict{str : value}, optional
        The dict of defines to apply to this preprocessed file if any;
        equivalent to the list of -DXYZ=ABC in the standard preprocessor
    includes : list of str, optional
        List of directory names to search for included files, if any
    notfound_fatal : bool, optional
        If True, include files not found are fatal. If False, they will simply
        be skipped (with a warning emitted). Default True

    Notes
    -----
    If ``fname`` is a file name, the directory containing that file name is the
    first directory searched for include files. If ``fname`` is a file-like
    object, then the current directory is the first searched.
    """

    def __init__(self, fname, defines=None, includes=None, notfound_fatal=True):
        if isinstance(fname, string_types):
            self._fileobj = genopen(fname, 'r')
            self._ownhandle = True
            curpath = path.abspath(path.split(fname)[0])
            self.filename = fname
        else:
            self._fileobj = fname
            self._ownhandle = False
            curpath = path.abspath(path.curdir)
            self.filename = None
        if includes is None:
            self._includes = [curpath]
        else:
            self._includes = [curpath] + list(includes)
        if defines is None:
            self.defines = OrderedDict()
        else:
            # Convert every define to a string
            self.defines = OrderedDict()
            for define, value in iteritems(defines):
                self.defines[define] = str(value)
        self._notfound_fatal = notfound_fatal

        # Now to keep track of other basic logic stuff
        self.included_files = []
        self._ifstack = []
        self._elsestack = []
        self._satisfiedstack = []
        self._num_ignoring_if = 0
        self._includefile = None

    def __del__(self):
        self.close()

    def close(self):
        if hasattr(self, '_ownhandle') and self._ownhandle:
            self._fileobj.close()

    def readline(self):
        try:
            return next(iter(self))
        except StopIteration:
            return ''

    def readlines(self):
        return [line for line in self]

    def read(self):
        return ''.join(self.readlines())

    def tell(self):
        return self._fileobj.tell()

    def seek(self, value):
        raise NotImplementedError('Cannot seek through a preprocessed file')

    def __iter__(self):
        """ Step through the preprocessed file line-by-line """
        for line in self._fileobj:
            rematch = ppre.match(line)
            if rematch:
                cmd, args = rematch.groups()
                args = args or ''
                self._ppcmdmap[cmd](self, args)
                # If we defined an include file, step through it
                if self._includefile is not None:
                    for line in self._includefile:
                        yield line
                    self._includefile.close()
                    # We have to pass our defines back to our caller
                    self.defines = self._includefile.defines
                    self._includefile = None
                continue

            if self._satisfiedstack and not self._satisfiedstack[-1]:
                # We are inside an unsatisfied conditional
                continue

            yield _replace_defines(line, self.defines)
        # Make sure we don't have any dangling ifs
        if self._ifstack:
            raise PreProcessorError('EOF: Unterminated #if(def)')

    @_strip_pp_comments
    def _pp_if(self, args):
        if self._satisfiedstack and not self._satisfiedstack[-1]:
            self._num_ignoring_if += 1
            return
        args = args.strip()
        if args.strip() in ('0', '1'):
            satisfied = bool(int(args))
            self._ifstack.append(str(bool(satisfied)))
            self._satisfiedstack.append(bool(satisfied))
            self._elsestack.append(_IN_IF)
            return
        raise NotImplementedError('Only "#if 0|1" is currently supported')

    @_strip_pp_comments
    def _pp_ifdef(self, args):
        if self._satisfiedstack and not self._satisfiedstack[-1]:
            self._num_ignoring_if += 1
            return
        words = args.split()
        if len(words) == 0:
            raise PreProcessorError('Bad #ifdef syntax: "#ifdef %s"' % args)
        elif len(words) > 1:
            warnings.warn('Ignored tokens in #ifdef: %s' % ', '.join(words[1:]),
                          PreProcessorWarning)
        self._ifstack.append('%s in self.defines' % words[0])
        self._elsestack.append(_IN_IF)
        self._satisfiedstack.append(words[0] in self.defines)

    @_strip_pp_comments
    def _pp_ifndef(self, args):
        if self._satisfiedstack and not self._satisfiedstack[-1]:
            self._num_ignoring_if += 1
            return
        words = args.split()
        if len(words) == 0:
            raise PreProcessorError('Bad #ifndef syntax: "#ifndef %s"' % args)
        elif len(words) > 1:
            warnings.warn('Ignored tokens in #ifndef: %s' % ', '.join(words[1:]),
                          PreProcessorWarning)
        self._ifstack.append('%s not in self.defines' % words[0])
        self._elsestack.append(_IN_IF)
        self._satisfiedstack.append(words[0] not in self.defines)

    @_strip_pp_comments
    def _pp_elif(self, args):
        raise NotImplementedError('#elif conditionals are not yet implemented.')

    @_strip_pp_comments
    def _pp_else(self, args):
        if self._num_ignoring_if > 0:
            return
        if not self._ifstack:
            raise PreProcessorError('#else missing #if(def)')
        words = args.split()
        if len(words) > 0:
            warnings.warn('Ignored tokens in #else: %s' % ', '.join(words[1:]),
                          PreProcessorWarning)
        if self._elsestack[-1] == _IN_ELSE:
            raise PreProcessorError('#else following #else')
        self._elsestack[-1] = _IN_ELSE
        self._satisfiedstack[-1] = not self._satisfiedstack[-1]

    @_strip_pp_comments
    def _pp_endif(self, args):
        if self._num_ignoring_if > 0:
            self._num_ignoring_if -= 1
            return
        if args.strip():
            warnings.warn('Ignored tokens in #endif: %s' % args.strip(),
                          PreProcessorWarning)
        if not self._ifstack:
            raise PreProcessorError('#endif missing #if(def)')
        self._ifstack.pop()
        self._satisfiedstack.pop()
        self._elsestack.pop()

    @_strip_pp_comments
    def _pp_include(self, args):
        if self._satisfiedstack and not self._satisfiedstack[-1]:
            return
        # Locate the include file
        rematch = includere.match(args)
        if not rematch:
            raise PreProcessorError('Bad #include syntax')
        includefile = rematch.groups()[0]
        self.included_files.append(includefile)
        for folder in self._includes:
            testfile = path.join(folder, includefile)
            if path.isfile(testfile):
                break
        else:
            if self._notfound_fatal:
                raise PreProcessorError('Could not find %s' % includefile)
            warnings.warn('Could not find %s; skipping' % includefile, PreProcessorWarning)
            return
        self._includefile = CPreProcessor(testfile,
                                          defines=self.defines,
                                          includes=self._includes,
                                          notfound_fatal=self._notfound_fatal)

    @_strip_pp_comments
    def _pp_define(self, args):
        if self._satisfiedstack and not self._satisfiedstack[-1]:
            return
        # Define a new variable
        words = args.split()
        if len(words) == 0:
            raise PreProcessorError('Nothing defined in #define')
        # Warn about a double-define
        if words[0] in self.defines:
            warnings.warn('%s already defined; overwriting' % words[0],
                          PreProcessorWarning)
        if len(words) == 1:
            self.defines[words[0]] = '1'
        elif len(words) >= 2:
            self.defines[words[0]] = args[len(words[0]):].strip()

    @_strip_pp_comments
    def _pp_undef(self, args):
        if self._satisfiedstack and not self._satisfiedstack[-1]:
            return
        # Define a new variable
        words = args.split()
        if len(words) == 1:
            try:
                del self.defines[words[0]]
            except KeyError:
                # Undefining an undefined variable is a no-op
                pass
        elif len(words) > 1:
            warnings.warn('Ignored tokens in #undef: %s' % ', '.join(words[1:]),
                          PreProcessorWarning)
        elif len(words) == 0:
            raise PreProcessorError('Nothing defined in #undef')

    # Context manager protocol
    def __exit__(self, type, value, traceback):
        self.close()

    def __enter__(self):
        return self

    _ppcmdmap = {'if' : _pp_if, 'elif' : _pp_elif, 'ifdef' : _pp_ifdef,
                 'else' : _pp_else, 'define' : _pp_define, 'undef' : _pp_undef,
                 'include' : _pp_include, 'endif' : _pp_endif,
                 'ifndef' : _pp_ifndef}

if __name__ == '__main__':
    # Act as a stand-alone preprocessor
    import argparse
    import sys
    parser = argparse.ArgumentParser()
    parser.add_argument('-i', '--input-file', dest='input', metavar='FILE',
                required=True, help='''Input file to pre-process. Either a file
                name or, if '-' is given, from standard input.''')
    parser.add_argument('-o', '--output-file', dest='output', metavar='FILE',
                default=None, help='''Output file with preprocessed results.
                Default is standard output''')
    parser.add_argument('-D', dest='defines', metavar='VAR[=VAL]',
                action='append', help='''List of predefined variables to pass to
                the preprocessor. Default VAL is 1 when missing.''', default=[])
    parser.add_argument('-I', dest='includes', metavar='DIRECTORY',
                action='append', help='''List of include directories to search
                for included files''', default=[])

    opt = parser.parse_args()

    defines = OrderedDict()
    for define in opt.defines:
        if '=' in define:
            define, val = define.split('=')
        else:
            val = '1'
        defines[define] = val

    if opt.input == '-':
        f = sys.stdin
    else:
        f = opt.input
    pp = CPreProcessor(f, defines=defines, includes=opt.includes)
    if opt.output is None:
        output = sys.stdout
        own_handle = False
    else:
        output = genopen(opt.output, 'w')
        own_handle = True

    for line in pp:
        output.write(line)
    if own_handle:
        output.close()



"""
List of all pointers and constants used in the Amber topology file.

Can be used like:
   from parmed.constants import *
"""
from __future__ import division
from math import pi as _pi, sqrt as _sqrt, log10 as _log10, acos as _acos

__all__ = ['AMBER_ELECTROSTATIC', 'AMBER_POINTERS', 'NATOM', 'NTYPES', 'NBONH', 'MBONA', 'NTHETH',
           'MTHETA', 'NPHIH', 'MPHIA', 'NHPARM', 'NPARM', 'NEXT', 'NRES', 'NBONA', 'NTHETA',
           'NPHIA', 'NUMBND', 'NUMANG', 'NPTRA', 'NATYP', 'NPHB', 'IFPERT', 'NBPER', 'NGPER',
           'NDPER', 'MBPER', 'MGPER', 'MDPER', 'IFBOX', 'NMXRS', 'IFCAP', 'NUMEXTRA', 'NCOPY',
           'NNB', 'RAD_TO_DEG', 'DEG_TO_RAD', 'TRUNCATED_OCTAHEDRON_ANGLE']

AMBER_ELECTROSTATIC = 18.2223
CHARMM_ELECTROSTATIC = _sqrt(332.0716)

AMBER_POINTERS = """
NATOM  : total number of atoms
NTYPES : total number of distinct atom types
NBONH  : number of bonds containing hydrogen
MBONA  : number of bonds not containing hydrogen
NTHETH : number of angles containing hydrogen
MTHETA : number of angles not containing hydrogen
NPHIH  : number of dihedrals containing hydrogen
MPHIA  : number of dihedrals not containing hydrogen
NHPARM : currently not used
NPARM  : currently not used
NEXT   : number of excluded atoms
NRES   : number of residues
NBONA  : MBONA + number of constraint bonds
NTHETA : MTHETA + number of constraint angles
NPHIA  : MPHIA + number of constraint dihedrals
NUMBND : number of unique bond types
NUMANG : number of unique angle types
NPTRA  : number of unique dihedral types
NATYP  : number of atom types in parameter file, see SOLTY below
NPHB   : number of distinct 10-12 hydrogen bond pair types
IFPERT : set to 1 if perturbation info is to be read in
NBPER  : number of bonds to be perturbed
NGPER  : number of angles to be perturbed
NDPER  : number of dihedrals to be perturbed
MBPER  : number of bonds with atoms completely in perturbed group
MGPER  : number of angles with atoms completely in perturbed group
MDPER  : number of dihedrals with atoms completely in perturbed groups
IFBOX  : set to 1 if standard periodic box, 2 when truncated octahedral
NMXRS  : number of atoms in the largest residue
IFCAP  : set to 1 if the CAP option from edit was specified
NUMEXTRA: number of extra points
NCOPY  : Number of copies for advanded simulations
"""
# These global variables provide a more natural way of accessing
# the various pointers.  Most useful if they're loaded into the
# top-level namespace.
NATOM  = 0; NTYPES = 1; NBONH  = 2; MBONA  = 3; NTHETH = 4
MTHETA = 5; NPHIH  = 6; MPHIA  = 7; NHPARM = 8; NPARM  = 9
NEXT   = 10; NRES   = 11; NBONA  = 12; NTHETA = 13; NPHIA  = 14
NUMBND = 15; NUMANG = 16; NPTRA  = 17; NATYP  = 18; NPHB   = 19
IFPERT = 20; NBPER  = 21; NGPER  = 22; NDPER  = 23; MBPER  = 24
MGPER  = 25; MDPER  = 26; IFBOX  = 27; NMXRS  = 28; IFCAP  = 29
NUMEXTRA= 30; NCOPY  = 31

# An alias
NNB = NEXT

RAD_TO_DEG = 180.0 / _pi
DEG_TO_RAD = _pi / 180.0
TRUNCATED_OCTAHEDRON_ANGLE = _acos(-1/3) * 180 / _pi

# For use in floating point comparisons
TINY = 1.0e-8
SMALL = 1.0e-4
TINY_DIGITS = int(_log10(TINY) + 0.5)
SMALL_DIGITS = int(_log10(SMALL) + 0.5)

# For I/O
DEFAULT_ENCODING = 'UTF-8'



"""
Provides a class for reading CHARMM-style files. The key component to these
files is that the ! character is a comment character and everything after ! is
ignored.
"""
from parmed.utils.io import genopen

class CharmmFile(object):
    """
    A CHARMM file that recognizes the "!" character as a 'comment' token. It
    can be iterated over and generally treated like a file object, but only
    spits out strings that have been truncated at its first comment character.
    
    There is currently no way to recognize a ! as a _non_ comment character,
    since allowing an escape character does not seem to be common practice and
    would likely introduce negative performance implications.
    """

    def __init__(self, fname, mode='r'):
        if mode not in ('r', 'w'):
            raise ValueError('Cannot open CharmmFile with mode "%s"' % mode)
        if mode == 'r':
            self.status = 'OLD'
        else:
            self.status = 'NEW'
        self._handle = genopen(fname, mode)
        self.closed = False
        self.line_number = 0
        self.comment = ''

    def __enter__(self):
        self._handle.__enter__()
        return self

    def __exit__(self, *args):
        if not self.closed:
            self.close()

    def tell(self):
        return self._handle.tell()

    def seek(self, value):
        return self._handle.seek(value)

    def write(self, *args, **kwargs):
        return self._handle.write(*args, **kwargs)

    def __iter__(self):
        # Iterate over the file
        parts = []
        for line in self._handle:
            try:
                idx = line.index('!')
            except ValueError:
                # There is no comment...
                idx = None
                end = ''
                self.comment = ''
                if line.rstrip('\r\n').endswith('-'):
                    # Continuation
                    parts.append(line.rstrip('\r\n')[:-1]) # Skip the continuation character
                    continue
            else:
                # Lines with no comment cannot continue
                end = '\n'
                self.comment = line[idx:].rstrip()
            parts.append(line[:idx] + end)
            yield ' '.join(parts)
            # Reset parts
            parts = []

    def readline(self):
        self.line_number += 1
        line = self._handle.readline()
        parts = []
        while line:
            if line.rstrip('\r\n').endswith('-'):
                # Continuation
                parts.append(line.rstrip('\r\n')[:-1]) # Skip the continuation character
                line = self._handle.readline()
                self.line_number += 1
            else:
                parts.append(line)
                break # done with this line
        line = ' '.join(parts)
        try:
            idx = line.index('!')
            self.comment = line[idx:].rstrip()
            end = '\n'
        except ValueError:
            idx = None
            end = ''
            self.comment = ''
        return line[:idx] + end

    def readlines(self):
        return [line for line in self]

    def read(self):
        return ''.join(self.readlines())

    def close(self):
        self._handle.close()
        self.closed = True

    def rewind(self):
        """ Return to the beginning of the file """
        self._handle.seek(0)

    def __del__(self):
        try:
            self.closed or self._handle.close()
        except AttributeError:
            # It didn't make it out of the constructor
            pass

# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

class CharmmStreamFile(object):
    """
    The stream file is broken down into sections of commands delimited by the
    strings:
        read <section> <options>
        ....
        ....
        end
    This object provides iterators over those sections and a file-like API for
    dealing with the text.

    """
    def __init__(self, fname):
        self.lines = []
        self.comments = []
        with CharmmFile(fname, 'r') as f:
            for line in f:
                self.lines.append(line)
                self.comments.append(f.comment)
        self.line_number = 0

    def __iter__(self):
        return iter(self.lines)

    def rewind(self):
        """ Return to the beginning of the file """
        self.line_number = 0

    def next_section(self):
        """
        Fast-forwards the file to the next CHARMM command section

        Returns
        -------
        name, data, comments : str, list of str, list of str
            name is the line defining the section that's being returned, whereas
            data is a list of all lines in the section, and comments is a list
            of all comments (same size as data) for all those lines

        Notes
        -----
        The line pointer will be set to the line defining the section
        """
        lines = []
        comments = []
        while self.line_number < len(self.lines):
            line = self.lines[self.line_number].strip()
            comment = self.comments[self.line_number].strip()
            if line[:4].lower() == 'read':
                title = line.strip()
                self.line_number += 1
                line = self.lines[self.line_number]
                while line and not line.strip().lower().startswith('end'):
                    lines.append(line)
                    comments.append(comment)
                    self.line_number += 1
                    line = self.lines[self.line_number]
                    comment = self.comments[self.line_number]
                if line[:3].upper() == 'END':
                    lines.append(line)
                    comments.append(comment)
                return title, lines, comments
            self.line_number += 1
        # No sections left
        return None, None, None


    def __del__(self): 
        pass



"""
A package dealing with different file formats and automatic detection of those
formats
"""

__all__ = ['load_file', 'PDBFile', 'CIFFile', 'Mol2File', 'PSFFile', 'PQRFile', 'SDFFile']

from parmed.formats.registry import load_file
from parmed.formats.mol2 import Mol2File
from parmed.formats.pdb import PDBFile, CIFFile
from parmed.formats.pqr import PQRFile
from parmed.formats.psf import PSFFile
from parmed.formats.sdf import SDFFile

# Now let's modify structure.Structure and add our write methods from our
# various formats
from parmed.structure import Structure as _Structure
_Structure.write_pdb = PDBFile.write
_Structure.write_cif = CIFFile.write
_Structure.write_psf = PSFFile.write



"""
This module contains functionality relevant to loading and parsing GROMACS GRO
(coordinate) files and building a stripped-down Structure from it
"""
from __future__ import print_function, division, absolute_import

from contextlib import closing
from parmed.constants import TINY
from parmed.exceptions import GromacsError
from parmed.formats.registry import FileFormatType
from parmed.geometry import (box_vectors_to_lengths_and_angles,
                             box_lengths_and_angles_to_vectors,
                             reduce_box_vectors)
from parmed.periodic_table import AtomicNum, element_by_name, Mass
from parmed.structure import Structure
from parmed.topologyobjects import Atom, ExtraPoint
from parmed import unit as u
from parmed.utils.io import genopen
from parmed.utils.six import add_metaclass, string_types

class _AtomLineParser(object):
    """ Parses atom lines from GRO files """
    def __init__(self):
        self._digits = None
        self._pdeci = 0
        self._ndeci = 0

    def read(self, line):
        """ Reads a line

        Parameters
        ----------
        line : str
            A line with an atom record from a GRO file

        Returns
        -------
        atom, resname, resnum : Atom, str, int
            The Atom instance, residue name, and residue number containing the
            atom
        """
        resnum = int(line[:5])
        resname = line[5:10].strip()
        atomname = line[10:15].strip()
        elem = element_by_name(atomname)
        atomic_number = AtomicNum[elem]
        mass = Mass[elem]
        atnum = int(line[15:20])
        if atomic_number == 0:
            atom = ExtraPoint(name=atomname, number=atnum)
        else:
            atom = Atom(atomic_number=atomic_number, name=atomname,
                        number=atnum, mass=mass)
        if self._digits is None:
            self._pdeci = line.index('.', 20)
            self._ndeci = line.index('.', self._pdeci+1)
            self._digits = self._ndeci - self._pdeci
        atom.xx, atom.xy, atom.xz = (
                float(line[20+i*self._digits:20+(i+1)*self._digits])*10
                    for i in range(3)
        )
        wbeg = 20 + self._digits * 3
        wend = wbeg + self._digits
        if line[wbeg:wend].strip():
            atom.vx, atom.vy, atom.vz = (
                    float(line[wbeg+i*self._digits:wend+i*self._digits])*10
                    for i in range(3)
            )
        return atom, resname, resnum

@add_metaclass(FileFormatType)
class GromacsGroFile(object):
    """ Parses and writes Gromacs GRO files """
    #===================================================

    @staticmethod
    def id_format(filename):
        """ Identifies the file as a GROMACS GRO file

        Parameters
        ----------
        filename : str
            Name of the file to check if it is a Gromacs GRO file

        Returns
        -------
        is_fmt : bool
            If it is identified as a Gromacs GRO file, return True. False
            otherwise
        """
        with closing(genopen(filename)) as f:
            f.readline() # Title line
            try:
                int(f.readline().strip()) # number of atoms
            except ValueError:
                return False
            line = f.readline()
            try:
                int(line[:5])
                if not line[5:10].strip(): return False
                if not line[10:15].strip(): return False
                int(line[15:20])
                pdeci = [i for i, x in enumerate(line) if x == '.']
                ndeci = pdeci[1] - pdeci[0] - 5
                for i in range(1, 4):
                    wbeg = (pdeci[0]-4)+(5+ndeci)*(i-1)
                    wend = (pdeci[0]-4)+(5+ndeci)*i
                    float(line[wbeg:wend])
                i = 4
                wbeg = (pdeci[0]-4)+(5+ndeci)*(i-1)
                wend = (pdeci[0]-4)+(5+ndeci)*i
                if line[wbeg:wend].strip():
                    for i in range(4, 7):
                        wbeg = (pdeci[0]-4)+(5+ndeci)*(i-1)
                        wend = (pdeci[0]-4)+(5+ndeci)*i
                        float(line[wbeg:wend])
            except ValueError:
                return False
            return True

    #===================================================

    @staticmethod
    def parse(filename, skip_bonds=False):
        """ Parses a Gromacs GRO file

        Parameters
        ----------
        filename : str or file-like
            Name of the file or the GRO file object
        skip_bonds : bool, optional
            If True, skip trying to assign bonds. This can save substantial time
            when parsing large files with non-standard residue names. However,
            no bonds are assigned. This is OK if, for instance, the GRO file is
            being parsed simply for its coordinates. This will also reduce the
            accuracy of assigned atomic numbers for typical ions. Default is
            False.

        Returns
        -------
        struct : :class:`Structure`
            The Structure instance instantiated with *just* residues and atoms
            populated (with coordinates)
        """
        struct = Structure()
        if isinstance(filename, string_types):
            fileobj = genopen(filename, 'r')
            own_handle = True
        else:
            fileobj = filename
            own_handle = False
        try:
            # Ignore the title line
            fileobj.readline()
            try:
                natom = int(fileobj.readline().strip())
            except ValueError:
                raise GromacsError('Could not parse %s as GRO file' % filename)
            line_parser = _AtomLineParser()
            for i, line in enumerate(fileobj):
                if i == natom: break
                try:
                    atom, resname, resnum = line_parser.read(line)
                except (ValueError, IndexError):
                    raise GromacsError('Could not parse the atom record of '
                                       'GRO file %s' % filename)
                struct.add_atom(atom, resname, resnum)
            else:
                # If no box exists, the break did not hit, so line still
                # contains the last atom (which cannot be interpreted as a box).
                # This wipes out line (IFF fileobj reached the line)
                line = fileobj.readline()
                if i+1 != natom:
                    raise GromacsError('Truncated GRO file. Found %d of %d '
                                       'atoms' % (i+1, natom))
            # Get the box from the last line if it's present
            if line.strip():
                try:
                    box = [float(x) for x in line.split()]
                except ValueError:
                    raise GromacsError('Could not understand box line of GRO '
                                       'file %s' % filename)
                if len(box) == 3:
                    struct.box = [box[0]*10, box[1]*10, box[2]*10,
                                  90.0, 90.0, 90.0]
                elif len(box) == 9:
                    # Assume we have vectors
                    leng, ang = box_vectors_to_lengths_and_angles(
                                [box[0], box[3], box[4]]*u.nanometers,
                                [box[5], box[1], box[6]]*u.nanometers,
                                [box[7], box[8], box[2]]*u.nanometers)
                    a, b, c = leng.value_in_unit(u.angstroms)
                    alpha, beta, gamma = ang.value_in_unit(u.degrees)
                    struct.box = [a, b, c, alpha, beta, gamma]
        finally:
            if own_handle:
                fileobj.close()

        # Assign bonds (and improved element guesses)
        if not skip_bonds:
            struct.assign_bonds()

        return struct

    #===================================================

    @staticmethod
    def write(struct, dest, precision=3, nobox=False, combine=False):
        """ Write a Gromacs Topology File from a Structure

        Parameters
        ----------
        struct : :class:`Structure`
            The structure to write to a Gromacs GRO file (must have coordinates)
        dest : str or file-like
            The name of a file or a file object to write the Gromacs topology to
        precision : int, optional
            The number of decimal places to print in the coordinates. Default 3
        nobox : bool, optional
            If the system does not have a periodic box defined, and this option
            is True, no box will be written. If False, the periodic box will be
            defined to enclose the solute with 0.5 nm clearance on all sides. If
            periodic box dimensions *are* defined, this variable has no effect.
        combine : 'all', None, or list of iterables, optional
            Equivalent to the combine argument of the GromacsTopologyFile.write
            method. If None, system atom order may be changed to meet the need
            for contiguously bonded groups of atoms to be part of a single
            moleculetype. All other values leave the atom order unchanged.
            Default is None.
        """

        def _write_atom_line(atom, atid, resid, has_vels, dest, precision):
            varwidth = 5 + precision
            crdfmt = '%%%d.%df' % (varwidth, precision)
            velfmt = '%%%d.%df' % (varwidth, precision+1)
            dest.write('%5d%-5s%5s%5d' % (resid, atom.residue.name[:5],
                                          atom.name[:5], atid))
            dest.write((crdfmt % (atom.xx/10))[:varwidth])
            dest.write((crdfmt % (atom.xy/10))[:varwidth])
            dest.write((crdfmt % (atom.xz/10))[:varwidth])
            if has_vels:
                dest.write((velfmt % (atom.vx/10))[:varwidth])
                dest.write((velfmt % (atom.vy/10))[:varwidth])
                dest.write((velfmt % (atom.vz/10))[:varwidth])
            dest.write('\n')

        own_handle = False
        if isinstance(dest, string_types):
            dest = genopen(dest, 'w')
            own_handle = True
        elif not hasattr(dest, 'write'):
            raise TypeError('dest must be a file name or file-like object')

        dest.write('GROningen MAchine for Chemical Simulation\n')
        dest.write('%5d\n' % len(struct.atoms))
        has_vels = all(hasattr(a, 'vx') for a in struct.atoms)
        if combine != 'all':
            resid, atid = 0, 0
            # use struct.split to get residue order as per topology file
            split_struct = struct.split()
            n_mols = sum(len(mol[1]) for mol in split_struct)
            unused_atoms = list(struct.atoms)
            for molid in range(n_mols):
                # loop through molids so we can get the correct molecule
                # according to the order they appear
                molecule = [
                    mol[0] for mol in split_struct if molid in mol[1]][0]
                new_molecule = set()  # track atoms added
                last_found_atom = None  # track when gro and top diverge

                for residue in molecule.residues:
                    resid += 1
                    for atom in residue.atoms:
                        # for each atom in split topology get the first
                        # matching occurrence in the original structure
                        for original_atom in unused_atoms:
                            if atom.type == original_atom.type and \
                               atom.name == original_atom.name and \
                               atom.residue.name == original_atom.residue.name:

                                if last_found_atom is not None and \
                                   original_atom.idx != last_found_atom.idx + 1:
                                    # a rearrangement has occurred! Need to do
                                    # extra check that we've found the correct
                                    # original_atom
                                    if len(new_molecule.intersection(
                                            original_atom.bond_partners)) == 0:
                                        # original_atom must be bonded to at
                                        # least one atom in the molecule we
                                        # are currently writing otherwise find
                                        # next candidate
                                        continue

                                atid += 1
                                _write_atom_line(
                                    original_atom, atid % 100000,
                                    resid % 100000, has_vels, dest, precision)
                                new_molecule.add(original_atom)
                                last_found_atom = original_atom
                                unused_atoms.remove(original_atom)
                                break
                        else:
                            raise RuntimeError("Could not find %s" % atom)
        else:
            for atom in struct.atoms:
                resid = (atom.residue.idx + 1) % 100000
                atid = (atom.idx + 1) % 100000
                _write_atom_line(
                    atom, atid, resid, has_vels, dest, precision)
                
        # Box, in the weird format...
        if struct.box is not None:
            a, b, c = reduce_box_vectors(*box_lengths_and_angles_to_vectors(
                            *struct.box))
            if all([abs(x-90) < TINY for x in struct.box[3:]]):
                dest.write('%10.5f'*3 % (a[0]/10, b[1]/10, c[2]/10))
            else:
                dest.write('%10.5f'*9 % (a[0]/10, b[1]/10, c[2]/10, a[1]/10,
                           a[2]/10, b[0]/10, b[2]/10, c[0]/10, c[1]/10))
            dest.write('\n')
        elif not nobox and struct.atoms:
            # Find the extent of the molecule in all dimensions, and buffer it
            # by 5 A
            crds = struct.coordinates
            diff = (crds.max(axis=1) - crds.min(axis=1)) / 10 + 0.5
            dest.write('%10.5f'*3 % (diff[0], diff[1], diff[2]))
            dest.write('\n')
        if own_handle:
            dest.close()



import numpy as np

class Symmetry(object):
    """Symmetry operations

    Parameters
    ----------
    tensor : np.array, shape=(3*n_symmetry, 4)
    """
    def __init__(self, tensor):
        self.data = np.asarray(tensor, dtype='f8')
        if self.data.ndim != 2 or self.data.shape[1] != 4:
            raise ValueError('Tensor shape must be (3*n_symmetry, 4)')



"""
Physical quantities with units for dimensional analysis and automatic unit
conversion.
"""
from __future__ import division

__docformat__ = "epytext en"

__author__ = "Christopher M. Bruns"
__copyright__ = "Copyright 2010, Stanford University and Christopher M. Bruns"
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Christopher M. Bruns"
__email__ = "cmbruns@stanford.edu"

# This code is copied from the `simtk.unit` package distributed as a standalone
# package (https://pypi.python.org/pypi/simtk.unit/) and as part of OpenMM
# (https://simtk.org/home/openmm).

# When OpenMM can be imported, the unit package will be taken from there.
# Otherwise, the implementation here will be used. This way, the
# `parmed.unit` package can be used interchangeably with OpenMM

try:
    from simtk.unit import *
except ImportError:
    from parmed.unit.unit import Unit, is_unit
    from parmed.unit.quantity import Quantity, is_quantity
    from parmed.unit.unit_math import *
    from parmed.unit.unit_definitions import *
    from parmed.unit.constants import *

# Now create the AKMA unit system, which is in common use by a number of
# programs (like Amber and CHARMM). Most complicated thing to get is the time.
# Do it from the kJ->kcal conversion

_time_scale = (daltons * angstroms**2 /
               kilocalories_per_mole).conversion_factor_to(picoseconds**2)
akma_time_base_unit = BaseUnit(time_dimension, "akma_time_unit", "akma-s")
akma_time_base_unit.define_conversion_factor_to(picosecond_base_unit,
            1 / sqrt(_time_scale))
del _time_scale

akma_unit_system = UnitSystem([
        angstrom_base_unit,
        dalton_base_unit,
        akma_time_base_unit,
        elementary_charge_base_unit,
        kelvin_base_unit,
        mole_base_unit,
        radian_base_unit])



"""
This stores a list of arguments, tokenizing a string into a list of arguments.
"""
import logging
import warnings
from .exceptions import NoArgument, InputError

LOGGER = logging.getLogger(__name__)

class Argument(object):
    """ Is an argument """
    def __init__(self, string):
        """ Input string for the argument """
        self.string = string.strip()
        self.marked = False
        self.keep_quotes = False
        if len(string) == 0 or (len(string) == 1 and string in ("'", '"')):
            LOGGER.warning("Unclosed quotation detected! Ignoring lone quote")

    def lower(self):
        """ Make the argument lower-case """
        return self.string.lower()

    def isfloat(self):
        """ Determines if it can be a double """
        try:
            tmp = float(self.string)
            del tmp
            return True
        except ValueError:
            return False

    def isint(self):
        """ Determines if it can be an integer """
        try:
            tmp = int(self.string)
            del tmp
            return True
        except ValueError:
            return False

    def ismask(self):
        """ Determines if we can be an Amber mask or not """
        # The easiest way to do this is to just search for ":" or "@" in the
        # input string (or just the string *, which matches everything).
        return self.string == '*' or ':' in self.string or '@' in self.string

    # Define some casting behavior
    def __float__(self):
        self.marked = True
        return float(self.string)

    def __int__(self):
        self.marked = True
        return int(self.string)

    def __str__(self):
        self.marked = True
        if self.keep_quotes:
            return self.string

        # Strip leading and trailing quotes (only if they both exist)
        if self.string[0] == "'" and self.string[-1] == "'":
            return self.string[1:-1]
        elif self.string[0] == '"' and self.string[-1] == '"':
            return self.string[1:len(self.string)-1]

        return self.string

    def __eq__(self, other):
        """ String comparison """
        return self.string == str(other)

    def mark(self):
        """ Provide a way of forcibly marking an argument """
        self.marked = True

    def unmark(self):
        """ Provide a way of forcibly unmarking an argument """
        self.marked = False

#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

class ArgumentList(object):
    """ 
    List of arguments.  
    The arguments should be parsed in the following order to ensure it's done
    correctly:
        get_key_*
        get_next_int
        get_next_float
        get_next_mask
        get_next_string
    """
    def __init__(self, inp):
        """ Defines the string """
        if isinstance(inp, tuple) or isinstance(inp, list):
            _inp = [l for l in inp if l is not None and l != '']
            self.original = ' '.join([str(l) for l in _inp])
            self.tokenlist = [Argument(str(l)) for l in _inp]
        elif isinstance(inp, ArgumentList):
            self.tokenlist = []
            try:
                while True:
                    arg = Argument(inp.get_next_string())
                    self.tokenlist.append(arg)
            except NoArgument:
                pass
            self.original = ' '.join([str(a) for a in inp.tokenlist])
        else:
            # Treat as a string-like and pad with spaces
            self.original = ' %s ' % inp
            self.tokenize(self.original)

    def tokenize(self, instring):
        """ 
        Tokenizes the line, everything between quotes is a single token,
        including whitespace
        """
        import re
        tokenre = re.compile(r'''((?:\s+".*?"\s+?)|(?:\s+'.*?'\s+?)|(?:\S+))''')
        tokenlist = tokenre.findall(instring)
        # Make sure every non-whitespace character is consumed
        if len(tokenre.sub('', instring).strip()) > 0:
            warnings.warn("Not all of argument string were handled: [%s]" % 
                          tokenre.sub('', instring).strip())
        # Strip out whitespace in all arguments
        self.tokenlist = [Argument(token.strip()) for token in tokenlist]

    def __str__(self):
        """ Return the original input string """
        return self.original

    def get_next_int(self, optional=False, default=None):
        """ Returns the next unmarked integer in the token list """
        for token in self.tokenlist:
            if token.isint() and not token.marked:
                return int(token)
        if optional: return default
        raise NoArgument("Missing integer argument!")

    def get_next_float(self, optional=False, default=None):
        """ Returns the next unmarked float in the token list """
        for token in self.tokenlist:
            if token.isfloat() and not token.marked:
                return float(token)
        if optional: return default
        raise NoArgument("Missing floating point argument!")

    def get_next_mask(self, optional=False, default=None):
        """ Returns the next string as long as it matches a mask """
        for token in self.tokenlist:
            if token.marked: continue
            if token.ismask():
                return str(token)
        if optional: return default
        raise NoArgument("Missing Amber mask!")

    def get_next_string(self, optional=False, keep_quotes=False, default=None):
        """ Returns the next unmarked float in the token list """
        for token in self.tokenlist:
            if not token.marked:
                token.keep_quotes = keep_quotes
                return str(token)
        if optional: return default
        raise NoArgument("Missing string!")

    def unmarked(self):
        """ Returns all unmarked arguments as a list of strings """
        unmarked_tokens = []
        for token in self.tokenlist:
            if not token.marked:
                unmarked_tokens.append(token.string)
        return unmarked_tokens

    def _get_key_arg(self, key, argtype):
        """ Gets a key with a given argument type """
        for i, token in enumerate(self.tokenlist):
            # Skip all marked tokens. This allows us to allow users to specify
            # a token numerous times (useful if you want to enable an
            # arbitrary-length list of arguments
            if token.marked: continue
            if token == key:
                if self.tokenlist[i+1].marked:
                    raise RuntimeError("Expected token already marked! Should "
                                       "not happen. This means a buggy "
                                       "action...")
                token.marked = True
                try:
                    return argtype(self.tokenlist[i+1])
                except ValueError as err:
                    raise InputError(str(err))

        raise NoArgument('Catch me!')

    def get_key_int(self, key, default):
        """ Get an integer that follows a keyword """
        try:
            return self._get_key_arg(key, int)
        except NoArgument:
            return default

    def get_key_float(self, key, default):
        """ Get a float that follows a keyword """
        try:
            return self._get_key_arg(key, float)
        except NoArgument:
            return default

    def get_key_mask(self, key, default):
        """ Get a mask that follows a keyword """
        for i, token in enumerate(self.tokenlist):
            if token == key:
                if not self.tokenlist[i+1].ismask():
                    raise InputError("Expected mask to follow %s. Got %s "
                                     "instead" %
                                     (key, self.tokenlist[i+1].string))
        try:
            return self._get_key_arg(key, str)
        except NoArgument:
            return default

    def get_key_string(self, key, default):
        """ Get a string that follows a keyword """
        try:
            return self._get_key_arg(key, str)
        except NoArgument:
            return default

    def has_key(self, key, mark=True):
        """ See if a particular keyword is present (optionally mark it) """
        for token in self.tokenlist:
            if not token.marked and token == key:
                if mark: token.mark()
                return True
        return False



"""
The GUI components of xparmed
"""
from __future__ import division
from parmed.utils.six import iteritems
from parmed.utils.six.moves import range
from parmed.utils.six.moves.tkinter import *
import parmed.utils.six.moves.tkinter_messagebox as tkMessageBox
from parmed.tools.gui.guiactions import gui_action_dispatcher
from parmed.tools.gui._guiwidgets import MessageWindow

BPR = 3 # Buttons Per Row

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

class ParmedApp(Frame):
    """ Main Parmed App Window """
    skipped_actions = ('parmout', 'go', 'quit', 'help', 'setoverwrite', 'ls',
                       'interpolate', 'listparms', 'parm', 'source', 'cd')
    def __init__(self, master, amber_prmtop):
        """ Instantiates the window """
        import math
        from parmed.tools.actions import COMMANDMAP
        global BPR
        Frame.__init__(self, master)
        self.parm = amber_prmtop
        actions = {}
        for action, actioncls in iteritems(COMMANDMAP):
            action_name = actioncls.__name__
            # Skip actions that don't get buttons
            if action in self.skipped_actions: continue
            # Save outparm for later
            elif action == 'outparm':
                outparm_action = _Description(action_name, actioncls.__doc__)
                continue
            actions[action] = _Description(action_name, actioncls.__doc__)
        self.message_frame = MessageWindow(self)
        # Now make the action button frame and put a title on it
        action_frame = Frame(self)
        label = Label(action_frame,text='ParmEd Actions (right-click for help)')
        action_keys = actions.keys(); action_keys.sort()
        # Number of Items Per Column
        IPC = int(math.ceil(len(action_keys) / BPR))
        for i, action in enumerate(action_keys):
            button = ActionButton(action_frame, self.parm, self.message_frame,
                                  actions[action])
            button.grid(column=i//IPC, row=i%IPC+1, padx=10, pady=5, 
                        sticky=N+S+E+W)
        action_frame.grid(column=0, row=0, pady=10, padx=5, sticky=N+S+E+W)
        self.message_frame.grid(column=1,row=0,pady=10,padx=5,sticky=N+S+E+W)
        label.grid(column=0, row=0, columnspan=BPR, sticky=N+S+E+W)
        # Add the two final buttons
        self.outparm = OutparmButton(self, self.parm, self.message_frame,
                                     outparm_action)
        self.outparm.grid(column=0, row=1, padx=10, sticky=N+S+E+W)
        self.exit = Button(self,text='Exit xParmEd',command=self.master.destroy)
        self.exit.grid(column=1, row=1, padx=10, sticky=N+S+E+W)
        # Control expandability of root window
        self.columnconfigure(0, weight=1)
        self.columnconfigure(1, weight=5)
        self.rowconfigure(0, weight=1)
        self.rowconfigure(1, weight=0)
        # Control expandability of action frame
        for i in range(int(math.ceil(len(action_keys)/BPR)) + 1):
            action_frame.rowconfigure(i+1, weight=1)
        for i in range(BPR):
            action_frame.columnconfigure(i, weight=1)

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

class _Description(object):
    " This formats a description so it can be displayed in different places "

    def __init__(self, action_name, desc_string):
        self.desc_string = desc_string
        self.action_name = action_name


    def info_box(self, event=None):
        """ Display an info button """
        tkMessageBox.showinfo(title=self.action_name, 
                        message=str(' '.join(self.desc_string.split())).strip())

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

class ActionButton(Button):
    " A button to dispatch the command to a particular function in guiactions "
    def __init__(self, root, amber_prmtop, messages, action_desc):
        self.amber_prmtop = amber_prmtop
        self.action_desc = action_desc
        self.messages = messages
        Button.__init__(self, root, text=action_desc.action_name,
                        command=self.execute)
        self.bind("<Button-3>", self.action_desc.info_box)

    def execute(self):
        """ Sends the action name to the GUI dispatcher """
        gui_action_dispatcher(self.master, self.amber_prmtop, 
                              self.action_desc.action_name, self.messages)

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

class OutparmButton(ActionButton):
    " An ActionButton specifically for outparm "
    def __init__(self, root, amber_prmtop, messages, action_desc):
        ActionButton.__init__(self, root, amber_prmtop, messages, action_desc)
        self.config(text='Write a New Topology File Now (can be used any '
                         'number of times)')

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~



"""
Convert an OpenMM Topology into a Structure instance, optionally filling in
parameters from a System
"""
from __future__ import absolute_import, division, print_function

import warnings
from collections import defaultdict

import numpy as np

from .. import unit as u
from ..exceptions import OpenMMWarning
from ..formats import load_file
from ..geometry import box_vectors_to_lengths_and_angles
from ..periodic_table import Element
from ..structure import Structure
from ..topologyobjects import (Angle, AngleType, Atom, AtomType, Bond, BondType, Cmap, CmapType,
                               Dihedral, DihedralType, ExtraPoint, Improper, ImproperType,
                               NonbondedException, NonbondedExceptionType, RBTorsionType,
                               UreyBradley)
from ..utils.decorators import needs_openmm
from ..utils.six import integer_types, iteritems, string_types
from ..utils.six.moves import range

__all__ = ['load_topology']


@needs_openmm
def load_topology(topology, system=None, xyz=None, box=None, condense_atom_types=True):
    """
    Creates a :class:`parmed.structure.Structure` instance from an OpenMM
    Topology, optionally filling in parameters from a System

    Parameters
    ----------
    topology : :class:`simtk.openmm.app.Topology`
        The Topology instance with the list of atoms and bonds for this system
    system : :class:`simtk.openmm.System` or str, optional
        If provided, parameters from this System will be applied to the
        Structure. If a string is given, it will be interpreted as the file name
        of an XML-serialized System, and it will be deserialized into a System
        before used to supply parameters
    xyz : str or array of float
        Name of a file containing coordinate information or an array of
        coordinates. If file has unit cell information, it also uses that
        information unless ``box`` (below) is also specified
    box : array of 6 floats
        Unit cell dimensions
    condense_atom_types : bool, default=True
        If True, create unique atom types based on de-duplicating properties. If False,
        create one atom type for each atom in the system, even if its properties match
        an existing atom type.

    Returns
    -------
    struct : :class:`Structure <parmed.structure.Structure>`
        The structure from the provided topology

    Raises
    ------
    OpenMMWarning if parameters are found that cannot be interpreted or
    processed by ParmEd

    TypeError if there are any mismatches between the provided topology and
    system (e.g., they have different numbers of atoms)

    IOError if system is a string and it is not an existing file

    Notes
    -----
    Due to its flexibility with CustomForces, it is entirely possible that the
    functional form of the potential will be unknown to ParmEd. This function
    will try to use the energy expression to identify supported potential types
    that are implemented as CustomForce objects. In particular, quadratic
    improper torsions, when recognized, will be extracted.

    Other CustomForces, including the CustomNonbondedForce used to implement
    NBFIX (off-diagonal L-J modifications) and the 12-6-4 potential, will not be
    processed and will result in an unknown functional form.

    If an OpenMM Atom.id attribute is populated by a non-integer, it will be
    used to name the corresponding ParmEd AtomType object.
    """
    import simtk.openmm as mm
    struct = Structure()
    atommap = dict()
    for c in topology.chains():
        chain = c.id
        for r in c.residues():
            residue = r.name
            resid = r.index
            for a in r.atoms():
                if a.element is None:
                    atom = ExtraPoint(name=a.name)
                else:
                    try:
                        aid = int(a.id)
                    except ValueError:
                        aid = a.id
                    atype = aid if not isinstance(aid, integer_types) else ''
                    atom = Atom(atomic_number=a.element.atomic_number,
                                name=a.name, mass=a.element.mass, type=atype)
                struct.add_atom(atom, residue, resid, chain)
                atommap[a] = atom
    for a1, a2 in topology.bonds():
        struct.bonds.append(Bond(atommap[a1], atommap[a2]))

    vectors = topology.getPeriodicBoxVectors()
    if vectors is not None:
        leng, ang = box_vectors_to_lengths_and_angles(*vectors)
        leng = leng.value_in_unit(u.angstroms)
        ang = ang.value_in_unit(u.degrees)
        struct.box = [leng[0], leng[1], leng[2], ang[0], ang[1], ang[2]]

    loaded_box = False

    if xyz is not None:
        if isinstance(xyz, string_types):
            xyz = load_file(xyz, skip_bonds=True)
            struct.coordinates = xyz.coordinates
            if struct.box is not None:
                if xyz.box is not None:
                    loaded_box = True
                    struct.box = xyz.box
        else:
            struct.coordinates = xyz

    if box is not None:
        loaded_box = True
        struct.box = box

    if struct.box is not None:
        struct.box = np.asarray(struct.box)

    if system is None:
        return struct

    if isinstance(system, string_types):
        system = load_file(system)

    if not isinstance(system, mm.System):
        raise TypeError('system must be an OpenMM System object or serialized '
                        'XML of an OpenMM System object')

    # We have a system, try to extract parameters from it
    if len(struct.atoms) != system.getNumParticles():
        raise TypeError('Topology and System have different numbers of atoms (%d vs. %d)' %
                        (len(struct.atoms), system.getNumParticles()))

    processed_forces = set()
    ignored_forces = (mm.CMMotionRemover, mm.AndersenThermostat, mm.MonteCarloBarostat,
                      mm.MonteCarloAnisotropicBarostat, mm.MonteCarloMembraneBarostat,
                      mm.CustomExternalForce, mm.GBSAOBCForce, mm.CustomGBForce)

    if system.usesPeriodicBoundaryConditions():
        if not loaded_box:
            vectors = system.getDefaultPeriodicBoxVectors()
            leng, ang = box_vectors_to_lengths_and_angles(*vectors)
            leng = leng.value_in_unit(u.angstroms)
            ang = ang.value_in_unit(u.degrees)
            struct.box = np.asarray([leng[0], leng[1], leng[2], ang[0], ang[1], ang[2]])
    else:
        struct.box = None

    for force in system.getForces():
        if isinstance(force, mm.HarmonicBondForce):
            if mm.HarmonicBondForce in processed_forces:
                # Try to process this HarmonicBondForce as a Urey-Bradley term
                _process_urey_bradley(struct, force)
            else:
                _process_bond(struct, force)
        elif isinstance(force, mm.HarmonicAngleForce):
            _process_angle(struct, force)
        elif isinstance(force, mm.PeriodicTorsionForce):
            _process_dihedral(struct, force)
        elif isinstance(force, mm.RBTorsionForce):
            _process_rbtorsion(struct, force)
        elif isinstance(force, mm.CustomTorsionForce):
            if not _process_improper(struct, force):
                struct.unknown_functional = True
                warnings.warn('Unknown functional form of CustomTorsionForce', OpenMMWarning)
        elif isinstance(force, mm.CMAPTorsionForce):
            _process_cmap(struct, force)
        elif isinstance(force, mm.NonbondedForce):
            _process_nonbonded(struct, force, condense_atom_types)
        elif isinstance(force, ignored_forces):
            continue
        else:
            struct.unknown_functional = True
            warnings.warn('Unsupported Force type %s' % type(force).__name__, OpenMMWarning)
        processed_forces.add(type(force))

    return struct

def _process_bond(struct, force):
    """ Adds bond parameters to the structure """
    typemap = dict()
    for ii in range(force.getNumBonds()):
        i, j, req, k = force.getBondParameters(ii)
        ai, aj = struct.atoms[i], struct.atoms[j]
        key = (req._value, k._value)
        if key in typemap:
            bond_type = typemap[key]
        else:
            bond_type = BondType(k*0.5, req)
            typemap[key] = bond_type
            struct.bond_types.append(bond_type)
        if aj in ai.bond_partners:
            for bond in ai.bonds:
                if aj in bond:
                    break
            else:
                raise RuntimeError('aj in ai.bond_partners, but couldn\'t find '
                                   'that bond!')
            bond.type = bond_type
        else:
            struct.bonds.append(Bond(ai, aj, type=bond_type))
    struct.bond_types.claim()

def _process_angle(struct, force):
    """ Adds angle parameters to the structure """
    typemap = dict()
    for ii in range(force.getNumAngles()):
        i, j, k, theteq, frc_k = force.getAngleParameters(ii)
        key = (theteq._value, frc_k._value)
        ai, aj, ak = struct.atoms[i], struct.atoms[j], struct.atoms[k]
        if key in typemap:
            angle_type = typemap[key]
        else:
            angle_type = AngleType(frc_k*0.5, theteq)
            typemap[key] = angle_type
            struct.angle_types.append(angle_type)
        struct.angles.append(Angle(ai, aj, ak, type=angle_type))
    struct.angle_types.claim()

def _process_urey_bradley(struct, force):
    """ Adds Urey-Bradley parameters to the structure """
    if not struct.angles:
        warnings.warn('Adding what seems to be Urey-Bradley terms before ' # pragma: no cover
                      'Angles. This is unexpected, but the parameters will '
                      'all be present in one form or another.', OpenMMWarning)
    typemap = dict()
    for ii in range(force.getNumBonds()):
        i, j, req, k = force.getBondParameters(ii)
        ai, aj = struct.atoms[i], struct.atoms[j]
        key = (req._value, k._value)
        if struct.angles and ai not in aj.angle_partners:
            warnings.warn('Adding what seems to be Urey-Bradley terms, but ' # pragma: no cover
                          'atoms %d and %d do not appear to be angled to each '
                          'other. Parameters will all be present, but may not '
                          'be in expected places.' % (ai.idx, aj.idx),
                          OpenMMWarning)
        if key in typemap:
            urey_type = typemap[key]
        else:
            urey_type = BondType(k*0.5, req)
            typemap[key] = urey_type
            struct.urey_bradley_types.append(urey_type)
        struct.urey_bradleys.append(UreyBradley(ai, aj, type=urey_type))
    struct.urey_bradley_types.claim()

def _process_dihedral(struct, force):
    """ Adds periodic torsions to the structure """
    typemap = dict()
    for ii in range(force.getNumTorsions()):
        i, j, k, l, per, phase, phi_k = force.getTorsionParameters(ii)
        ai, aj = struct.atoms[i], struct.atoms[j]
        ak, al = struct.atoms[k], struct.atoms[l]
        key = (per, phase._value, phi_k._value)
        if key in typemap:
            dihed_type = typemap[key]
        else:
            dihed_type = DihedralType(phi_k, per, phase)
            typemap[key] = dihed_type
            struct.dihedral_types.append(dihed_type)
        improper = (ai in ak.bond_partners and aj in ak.bond_partners and
                    al in ak.bond_partners)
        struct.dihedrals.append(Dihedral(ai, aj, ak, al, improper=improper,
                                         type=dihed_type))
    struct.dihedral_types.claim()

def _process_rbtorsion(struct, force):
    """ Adds Ryckaert-Bellemans torsions to the structure """
    typemap = dict()
    for ii in range(force.getNumTorsions()):
        i, j, k, l, c0, c1, c2, c3, c4, c5 = force.getTorsionParameters(ii)
        ai, aj = struct.atoms[i], struct.atoms[j]
        ak, al = struct.atoms[k], struct.atoms[l]
        # TODO -- Fix this when OpenMM is fixed
        try:
            key = (c0._value, c1._value, c2._value, c3._value, c4._value, c5._value)
            f = 1                          # pragma: no cover
        except AttributeError:             # pragma: no cover
            key = (c0, c1, c2, c3, c4, c5) # pragma: no cover
            f = u.kilojoules_per_mole      # pragma: no cover
        if key in typemap:
            dihed_type = typemap[key]
        else:
            dihed_type = RBTorsionType(c0*f, c1*f, c2*f, c3*f, c4*f, c5*f)
            typemap[key] = dihed_type
            struct.rb_torsion_types.append(dihed_type)
        struct.rb_torsions.append(Dihedral(ai, aj, ak, al, type=dihed_type))
    struct.rb_torsion_types.claim()

def _process_improper(struct, force):
    """ Processes a CustomTorsionForce and looks at the energy expression to see
    if it's a quadratic improper torsion. Then adds the parameters if applicable

    Returns
    -------
    is_improper : bool
        Returns True if the energy expression is recognized as a quadratic
        improper, and False otherwise
    """
    eqn = force.getEnergyFunction().replace(' ', '')
    if ';' in eqn: # Just look at the first segment of the equation if there are multiple
        eqn = eqn[:eqn.index(';')]
    # Don't try to be fancy with regexes for fear of making a possible mistake.
    # ParmEd and OpenMM use only these two eqns for the improper torsions:
    #  k*(theta-theta0)^2 vs. 0.5*k*(theta-theta0)^2
    # So only recognize the above 2 forms
    if eqn not in ('0.5*k*(theta-theta0)^2', 'k*(theta-theta0)^2', 'k*dtheta_torus^2',
                   '0.5*k*dtheta_torus^2'):
        return False
    if eqn.startswith('0.5'):
        fac = 0.5
    else:
        fac = 1
    typemap = dict()
    for ii in range(force.getNumTorsions()):
        # All formulations put k first and equilibrium angle second, in radians
        i, j, k, l, (psi_k, psi_eq) = force.getTorsionParameters(ii)
        ai, aj = struct.atoms[i], struct.atoms[j]
        ak, al = struct.atoms[k], struct.atoms[l]
        key = (psi_k, psi_eq)
        if key in typemap:
            imp_type = typemap[key]
        else:
            imp_type = ImproperType(psi_k*fac*u.kilojoule_per_mole/u.radian**2,
                                    psi_eq*u.radian)
            typemap[key] = imp_type
            struct.improper_types.append(imp_type)
        struct.impropers.append(Improper(ai, aj, ak, al, type=imp_type))
    struct.improper_types.claim()
    return True

def _process_cmap(struct, force):
    """ Adds CMAPs to the structure """
    # store the list of cmap types
    cmap_types = []
    for ii in range(force.getNumMaps()):
        size, grid = force.getMapParameters(ii)
        # Future-proof in case units start getting added to these maps
        if u.is_quantity(grid):
            typ = CmapType(size, grid)                       # pragma: no cover
        else:
            typ = CmapType(size, grid*u.kilojoules_per_mole) # pragma: no cover
        cmap_types.append(typ)
        typ.grid = typ.grid.T.switch_range()
        typ.used = False
    # Add all cmaps
    for ii in range(force.getNumTorsions()):
        mapidx, ii, ij, ik, il, ji, jj, jk, jl = force.getTorsionParameters(ii)
        if ij != ji or ik != jj or il != jk:
            warnings.warn('Non-continuous CMAP torsions detected. Not ' # pragma: no cover
                          'supported.', OpenMMWarning)
            continue # pragma: no cover
        ai, aj, ak = struct.atoms[ii], struct.atoms[ij], struct.atoms[ik]
        al, am = struct.atoms[il], struct.atoms[jl]
        cmap_type = cmap_types[mapidx]
        cmap_type.used = True
        struct.cmaps.append(Cmap(ai, aj, ak, al, am, type=cmap_type))
    for cmap_type in cmap_types:
        if cmap_type.used:
            struct.cmap_types.append(cmap_type)
    struct.cmap_types.claim()

def _process_nonbonded(struct, force, condense_atom_types):
    """ Adds nonbonded parameters to the structure """
    typemap = dict()
    element_typemap = defaultdict(int)
    assert force.getNumParticles() == len(struct.atoms), "Atom # mismatch"
    for i in range(force.getNumParticles()):
        atom = struct.atoms[i]
        chg, sig, eps = force.getParticleParameters(i)
        atype_name = (atom.type if atom.type != ''
                      else Element[atom.atomic_number])
        key = (atype_name, sig._value, eps._value)
        if key in typemap and condense_atom_types:
            atom_type = typemap[key]
        else:
            if atom.type == '':
                element_typemap[atype_name] += 1
                atype_name = '%s%d' % (atype_name, element_typemap[atype_name])
            typemap[key] = atom_type = AtomType(atype_name, None, atom.mass,
                                                atom.atomic_number)
        atom.charge = chg.value_in_unit(u.elementary_charge)
        rmin = sig.value_in_unit(u.angstroms) * 2**(1/6) / 2 # to rmin/2
        eps = eps.value_in_unit(u.kilocalories_per_mole)
        atom_type.set_lj_params(eps, rmin)
        atom.atom_type = atom_type
        atom.type = atom_type.name

    explicit_exceptions = defaultdict(set)
    bond_graph_exceptions = defaultdict(set)
    for atom in struct.atoms:
        for a2 in atom.bond_partners:
            if atom is not a2:
                bond_graph_exceptions[atom].add(a2)
            for a3 in a2.bond_partners:
                if a3 is atom: continue
                if atom is not a3:
                    bond_graph_exceptions[atom].add(a3)

    # TODO should we compress exception types?
    for ii in range(force.getNumExceptions()):
        i, j, q, sig, eps = force.getExceptionParameters(ii)
        q = q.value_in_unit(u.elementary_charge**2)
        sig = sig.value_in_unit(u.angstrom)
        eps = eps.value_in_unit(u.kilocalorie_per_mole)
        ai, aj = struct.atoms[i], struct.atoms[j]
        if q == 0 and (sig == 0 or eps == 0):
            explicit_exceptions[ai].add(aj)
            explicit_exceptions[aj].add(ai)
            continue
        try:
            chgscale = q / (ai.charge * aj.charge)
        except ZeroDivisionError:
            if q != 0:
                raise ValueError("Can't scale charge product 0 to match %s" % q)
            chgscale = None
        nbtype = NonbondedExceptionType(sig*2**(1/6), eps, chgscale)
        struct.adjusts.append(NonbondedException(ai, aj, type=nbtype))
        struct.adjust_types.append(nbtype)
    struct.adjust_types.claim()
    # Go through all adjust_types and replace any chgscale values that are None with the first
    # non-None value present. If all are None, set all to 1.0. This way we maximize the likelihood
    # that all generated systems have 1 scaling factor which makes it easier to translate to other
    # formats that may not support multiple scaling factors in electrostatic scaling (like GROMACS)
    first_scaling_factor = 1.0
    for adjust_type in struct.adjust_types:
        if adjust_type.chgscale is not None:
            first_scaling_factor = adjust_type.chgscale
            break
    # Now go through and set all Nones to first_scaling_factor
    for adjust_type in struct.adjust_types:
        if adjust_type.chgscale is None:
            adjust_type.chgscale = first_scaling_factor

    # Check that all of our exceptions are accounted for
    for ai, exceptions in iteritems(bond_graph_exceptions):
        if exceptions - explicit_exceptions[ai]:
            struct.unknown_functional = True
            warnings.warn('Detected incomplete exceptions. Not supported.', OpenMMWarning)
            break



""" A list of all of the tools used by the parmed script """

__author__ = 'Jason Swails'
__all__ = [] # This is populated with the ParmEd Actions below

from parmed.tools.actions import *
from parmed.tools.parmlist import ParmList

__all__ += ['ParmList']



"""
This module contains classes for parsing and processing CHARMM parameter,
topology, and stream files. It only extracts atom properties from the
topology files and extracts all parameters from the parameter files

Author: Jason M. Swails
"""
from __future__ import absolute_import, division, print_function

import warnings
from collections import OrderedDict
from copy import copy
from itertools import permutations

from .exceptions import ParameterError, ParameterWarning
from .topologyobjects import (AtomType, DihedralType, DihedralTypeList,
                              NoUreyBradley, UnassignedAtomType)
from .utils.six import iteritems, itervalues
from .utils.six.moves import range #pylint: disable=W0622,E0401

class ParameterSet(object):
    """
    Stores a parameter set defining a force field

    Attributes
    ----------
    atom_types : dict(str:AtomType)
        Dictionary mapping the names of the atom types to the corresponding AtomType instances
    atom_types_int : dict(int:AtomType)
        Dictionary mapping the serial indexes of the atom types to the corresponding AtomType
        instances
    atom_types_tuple : dict((str,int):AtomType)
        Dictionary mapping the (name,number) tuple of the atom types to the corresponding AtomType
        instances
    bond_types : dict((str,str):AtomType)
        Dictionary mapping the 2-element tuple of the names of the two atom types involved in the
        bond to the BondType instances
    angle_types : dict((str,str,str):AngleType)
        Dictionary mapping the 3-element tuple of the names of the three atom types involved in the
        angle to the AngleType instances
    urey_bradley_types : dict((str,str,str):BondType)
        Dictionary mapping the 3-element tuple of the names of the three atom types involved in the
        angle to the BondType instances of the Urey-Bradley terms
    dihedral_types : dict((str,str,str,str):list(DihedralType))
        Dictionary mapping the 4-element tuple of the names of the four atom types involved in the
        dihedral to the DihedralType instances. Since each torsion term can be a multiterm
        expansion, each item corresponding to a key in this dict is a list of `DihedralType`s for
        each term in the expansion
    improper_types : dict((str,str,str,str):ImproperType)
        Dictionary mapping the 4-element tuple of the names of the four atom types involved in the
        improper torsion to the ImproperType instances
    improper_periodic_types : dict((str,str,str,str):DihedralType)
        Dictionary mapping the 4-element tuple of the names of the four atom types involved in the
        improper torsion (modeled as a Fourier series) to the DihedralType instances. Note, the
        central atom should always be put in the *third* position of the key
    rb_torsion_types : dict((str,str,str,str):RBTorsionType)
        Dictionary mapping the 4-element tuple of the names of the four atom types involved in the
        Ryckaert-Bellemans torsion to the RBTorsionType instances
    cmap_types : dict((str,str,str,str,str,str,str,str):CmapType)
        Dictionary mapping the 5-element tuple of the names of the five atom types involved in the
        correction map to the CmapType instances
    nbfix_types : dict((str,str):(float,float))
        Dictionary mapping the 2-element tuple of the names of the two atom types whose LJ terms are
        modified to the tuple of the (epsilon,rmin) terms for that off-diagonal term
    pair_types : dict((str,str):NonbondedExceptionType)
        Dictionary mapping the 2-element tuple of atom type names for which explicit exclusion rules
        should be applied
    parametersets : list(str)
        List of parameter set names processed in the current ParameterSet
    residues : dict(str:ResidueTemplate|ResidueTemplateContainer)
        A library of ResidueTemplate objects mapped to the residue name defined in the force field
        library files
    """

    def __init__(self):
        # Instantiate the list types
        self.atom_types = self.atom_types_str = OrderedDict()
        self.atom_types_int = OrderedDict()
        self.atom_types_tuple = OrderedDict()
        self.bond_types = OrderedDict()
        self.angle_types = OrderedDict()
        self.urey_bradley_types = OrderedDict()
        self.dihedral_types = OrderedDict()
        self.improper_types = OrderedDict()
        self.improper_periodic_types = OrderedDict()
        self.rb_torsion_types = OrderedDict()
        self.cmap_types = OrderedDict()
        self.nbfix_types = OrderedDict()
        self.pair_types = OrderedDict()
        self.parametersets = []
        self._combining_rule = 'lorentz'
        self.residues = OrderedDict()
        self.patches = OrderedDict()
        self.default_scee = self.default_scnb = 1.0
        self._improper_key_map = OrderedDict()

    def __copy__(self):
        other = type(self)()
        for key, item in iteritems(self.atom_types):
            other.atom_types[key] = copy(item)
        for key, item in iteritems(self.atom_types_int):
            other.atom_types_int[key] = copy(item)
        for key, item in iteritems(self.atom_types_tuple):
            other.atom_types_tuple[key] = copy(item)
        for key, item in iteritems(self.bond_types):
            if key in other.bond_types: continue
            typ = copy(item)
            other.bond_types[key] = typ
            other.bond_types[tuple(reversed(key))] = typ
        for key, item in iteritems(self.pair_types):
            if key in other.pair_types: continue
            typ = copy(item)
            other.pair_types[key] = typ
            other.pair_types[tuple(reversed(key))] = typ
        for key, item in iteritems(self.angle_types):
            if key in other.angle_types: continue
            typ = copy(item)
            other.angle_types[key] = typ
            other.angle_types[tuple(reversed(key))] = typ
        for key, item in iteritems(self.dihedral_types):
            if key in other.dihedral_types: continue
            typ = copy(item)
            other.dihedral_types[key] = typ
            other.dihedral_types[tuple(reversed(key))] = typ
        for key, item in iteritems(self.rb_torsion_types):
            if key in other.rb_torsion_types: continue
            typ = copy(item)
            other.rb_torsion_types[key] = typ
            other.rb_torsion_types[tuple(reversed(key))] = typ
        for key, item in iteritems(self.improper_types):
            if key in other.improper_types: continue
            other.improper_types[key] = copy(item)
        for key, item in iteritems(self.improper_periodic_types):
            if key in other.improper_periodic_types: continue
            typ = copy(item)
            other.improper_periodic_types[key] = typ
            other.improper_periodic_types[tuple(reversed(key))] = typ
        for key, item in iteritems(self.urey_bradley_types):
            if key in other.urey_bradley_types: continue
            typ = copy(item)
            other.urey_bradley_types[key] = typ
            other.urey_bradley_types[tuple(reversed(key))] = typ
        for key, item in iteritems(self.cmap_types):
            if key in other.cmap_types: continue
            typ = copy(item)
            other.cmap_types[key] = typ
            other.cmap_types[tuple(reversed(key))] = typ
        for key, item in iteritems(self.residues):
            other.residues[key] = copy(item)
        for key, item in iteritems(self.patches):
            other.patches[key] = copy(item)
        other._improper_key_map = copy(self._improper_key_map)
        other.combining_rule = self.combining_rule

        return other

    @classmethod
    def from_structure(cls, struct, allow_unequal_duplicates=True):
        """ Extracts known parameters from a Structure instance

        Parameters
        ----------
        struct : :class:`parmed.structure.Structure`
            The parametrized ``Structure`` instance from which to extract
            parameters into a ParameterSet
        allow_unequal_duplicates : bool, optional
            If True, if two or more unequal parameter types are defined by the
            same atom types, the last one encountered will be assigned. If
            False, an exception will be raised. Default is True

        Returns
        -------
        params : :class:`ParameterSet`
            The parameter set with all parameters defined in the Structure

        Notes
        -----
        The parameters here are copies of the ones in the Structure, so
        modifying the generated ParameterSet will have no effect on ``struct``.
        Furthermore, the *first* occurrence of each parameter will be used. If
        future ones differ, they will be silently ignored, since this is
        expected behavior in some instances (like with Gromacs topologies in the
        ff99sb-ildn force field) unless ``allow_unequal_duplicates`` is set to
        ``False``

        Dihedrals are a little trickier. They can be multi-term, which can be
        represented either as a *single* entry in dihedrals with a type of
        DihedralTypeList or multiple entries in dihedrals with a DihedralType
        parameter type. In this case, the parameter is constructed from either
        the first DihedralTypeList found or the first DihedralType of each
        periodicity found if no matching DihedralTypeList is found.

        Raises
        ------
        :class:`parmed.exceptions.ParameterError` if allow_unequal_duplicates is
        False and 2+ unequal parameters are defined between the same atom types.

        `NotImplementedError` if any AMOEBA potential terms are defined in the
        input structure
        """
        params = cls()
        found_dihed_type_list = dict()
        for atom in struct.atoms:
            if atom.atom_type in (UnassignedAtomType, None):
                atom_type = AtomType(atom.type, None, atom.mass, atom.atomic_number)
                atom_type.set_lj_params(atom.epsilon, atom.rmin, atom.epsilon_14, atom.rmin_14)
                params.atom_types[atom.type] = atom_type
            else:
                atom_type = copy(atom.atom_type)
                params.atom_types[str(atom_type)] = atom_type
                if atom_type.number is not None:
                    params.atom_types_int[int(atom_type)] = atom_type
                    params.atom_types_tuple[(int(atom_type), str(atom_type))] = atom_type
        if struct.has_NBFIX():
            for atom in struct.atoms:
                if atom.atom_type.nbfix:
                    other_atoms = list(atom.atom_type.nbfix.keys())
                    for other_atom in other_atoms:
                        (rmin, epsilon, rmin14, epsilon14) = atom.atom_type.nbfix[other_atom]
                        if (other_atom, atom.type) in params.nbfix_types:
                            continue
                        params.nbfix_types[(atom.type, other_atom)] = (epsilon, rmin)
        for bond in struct.bonds:
            if bond.type is None: continue
            key = (bond.atom1.type, bond.atom2.type)
            if key in params.bond_types:
                if not allow_unequal_duplicates and params.bond_types[key] != bond.type:
                    raise ParameterError('Unequal bond types defined between %s and %s' % key)
                continue # pragma: no cover
            typ = copy(bond.type)
            key = (bond.atom1.type, bond.atom2.type)
            params.bond_types[key] = typ
            params.bond_types[tuple(reversed(key))] = typ
        for angle in struct.angles:
            if angle.type is None: continue
            key = (angle.atom1.type, angle.atom2.type, angle.atom3.type)
            if key in params.angle_types:
                if not allow_unequal_duplicates and params.angle_types[key] != angle.type:
                    raise ParameterError('Unequal angle types defined between %s, %s, and %s' % key)
                continue # pragma: no cover
            typ = copy(angle.type)
            key = (angle.atom1.type, angle.atom2.type, angle.atom3.type)
            params.angle_types[key] = typ
            params.angle_types[tuple(reversed(key))] = typ
            if angle.funct == 5:
                key = (angle.atom1.type, angle.atom2.type, angle.atom3.type)
                params.urey_bradley_types[key] = NoUreyBradley
                params.urey_bradley_types[tuple(reversed(key))] = NoUreyBradley
        for dihedral in struct.dihedrals:
            if dihedral.type is None: continue
            if dihedral.improper:
                # Make sure the central atom comes third, but add all permutations since it's not
                # clear which particular ordering the various programs require
                for key in _find_improper_keys(dihedral):
                    if key in params.improper_periodic_types:
                        if (not allow_unequal_duplicates and
                                params.improper_periodic_types[key] != dihedral.type):
                            raise ParameterError('Unequal dihedral types defined between %s, %s, '
                                                 '%s, and %s' % key)
                        continue # pragma: no cover
                    typ = copy(dihedral.type)
                    params.improper_periodic_types[key] = typ
            else:
                key = (dihedral.atom1.type, dihedral.atom2.type,
                       dihedral.atom3.type, dihedral.atom4.type)
                # Proper dihedral. Look out for multi-term forms
                if key in params.dihedral_types and found_dihed_type_list[key]:
                    # Already found a multi-term dihedral type list
                    if not allow_unequal_duplicates:
                        if isinstance(dihedral.type, DihedralTypeList):
                            if params.dihedral_types[key] != dihedral.type:
                                raise ParameterError('Unequal dihedral types defined between %s, '
                                                     '%s, %s, and %s' % key)
                        elif isinstance(dihedral.type, DihedralType):
                            for dt in params.dihedral_types[key]:
                                if dt == dihedral.type:
                                    break
                            else:
                                raise ParameterError('Unequal dihedral types defined between %s, '
                                                     '%s, %s, and %s' % key)
                    continue # pragma: no cover
                elif key in params.dihedral_types:
                    # We have one term of a potentially multi-term dihedral.
                    if isinstance(dihedral.type, DihedralTypeList):
                        # This is a full Fourier series list
                        found_dihed_type_list[key] = True
                        found_dihed_type_list[tuple(reversed(key))] = True
                        typ = copy(dihedral.type)
                        params.dihedral_types[key] = typ
                        params.dihedral_types[tuple(reversed(key))] = typ
                    else:
                        # This *might* be another term. Make sure another term
                        # with its periodicity does not already exist
                        for t in params.dihedral_types[key]:
                            if t.per == dihedral.type.per:
                                if not allow_unequal_duplicates and t != dihedral.type:
                                    raise ParameterError('Unequal dihedral types defined bewteen '
                                                         '%s, %s, %s, and %s' % key)
                                break
                        else:
                            # If we got here, we did NOT find this periodicity.
                            # And since this is mutating a list in-place, it
                            # automatically propagates to the reversed key
                            typ = copy(dihedral.type)
                            params.dihedral_types[key].append(typ)
                else:
                    # New parameter. If it's a DihedralTypeList, assign it and
                    # be done with it. If it's a DihedralType, start a
                    # DihedralTypeList to be added to later.
                    if isinstance(dihedral.type, DihedralTypeList):
                        found_dihed_type_list[key] = True
                        found_dihed_type_list[tuple(reversed(key))] = True
                        typ = copy(dihedral.type)
                        params.dihedral_types[key] = typ
                        params.dihedral_types[tuple(reversed(key))] = typ
                    else:
                        found_dihed_type_list[key] = False
                        found_dihed_type_list[tuple(reversed(key))] = False
                        typ = DihedralTypeList()
                        typ.append(copy(dihedral.type))
                        params.dihedral_types[key] = typ
                        params.dihedral_types[tuple(reversed(key))] = typ
        for improper in struct.impropers:
            if improper.type is None: continue
            key = (improper.atom1.type, improper.atom2.type,
                   improper.atom3.type, improper.atom4.type)
            if key in params.improper_types:
                if not allow_unequal_duplicates and params.improper_types[key] != improper.type:
                    raise ParameterError('Unequal improper types defined between '
                                         '%s, %s, %s, and %s' % key)
                continue # pragma: no cover
            params.improper_types[key] = copy(improper.type)
        for cmap in struct.cmaps:
            if cmap.type is None: continue
            key = (cmap.atom1.type, cmap.atom2.type, cmap.atom3.type, cmap.atom4.type,
                   cmap.atom2.type, cmap.atom3.type, cmap.atom4.type, cmap.atom5.type)
            if key in params.cmap_types:
                if not allow_unequal_duplicates and cmap.type != params.cmap_types[key]:
                    raise ParameterError('Unequal CMAP types defined between %s, %s, %s, %s, and '
                                         '%s' % (key[0], key[1], key[2], key[3], key[7]))
                continue # pragma: no cover
            typ = copy(cmap.type)
            params.cmap_types[key] = typ
            params.cmap_types[tuple(reversed(key))] = typ
        urey_brads_preassigned = len(params.urey_bradley_types) > 0
        for urey in struct.urey_bradleys:
            if urey.type is None or urey.type is NoUreyBradley: continue
            key = _find_ureybrad_key(urey)
            if key is None: continue
            if urey_brads_preassigned and key not in params.urey_bradley_types:
                warnings.warn('Angle corresponding to Urey-Bradley type not found',
                              ParameterWarning)
            typ = copy(urey.type)
            params.urey_bradley_types[key] = typ
            params.urey_bradley_types[tuple(reversed(key))] = typ
        if not urey_brads_preassigned and len(params.urey_bradley_types) > 0:
            # Go through all of our angle parameters and make sure there is a
            # matching Urey-Bradley list. If there's not, that means there is no
            # Urey-Bradley term for that angle
            for key in params.angle_types:
                if key in params.urey_bradley_types: continue
                params.urey_bradley_types[key] = NoUreyBradley
        for adjust in struct.adjusts:
            if adjust.type is None: continue
            key = (adjust.atom1.type, adjust.atom2.type)
            if key in params.pair_types:
                if not allow_unequal_duplicates and params.pair_types[key] != adjust.type:
                    raise ParameterError('Unequal pair types defined between %s and %s' % key)
                continue # pragma: no cover
            typ = copy(adjust.type)
            params.pair_types[key] = typ
            params.pair_types[tuple(reversed(key))] = typ
        # Trap for Amoeba potentials
        if (struct.trigonal_angles or struct.out_of_plane_bends or struct.torsion_torsions or
                struct.stretch_bends or struct.trigonal_angles or struct.pi_torsions):
            raise NotImplementedError('Cannot extract parameters from an Amoeba-parametrized '
                                      'system yet')
        return params

    def condense(self, do_dihedrals=True):
        """
        This function goes through each of the parameter type dicts and
        eliminates duplicate types. After calling this function, every unique
        bond, angle, dihedral, improper, or cmap type will pair with EVERY key
        in the type mapping dictionaries that points to the equivalent type

        Parameters
        ----------
        do_dihedrals : bool=True
            Dihedrals can take the longest time to compress since testing their
            equality takes the longest (this is complicated by the existence of
            multi-term torsions). This flag will allow you to *skip* condensing
            the dihedral parameter types (for large parameter sets, this can cut
            the compression time in half)

        Returns
        -------
        self
            The instance that is being condensed

        Notes
        -----
        The return value allows you to condense the types at construction time.

        Example
        -------
        >>> params = ParameterSet().condense()
        >>> params
        <parmed.parameters.ParameterSet at 0x7f88757de090>
        """
        # First scan through all of the bond types
        self._condense_types(self.bond_types)
        self._condense_types(self.angle_types)
        self._condense_types(self.urey_bradley_types)
        if do_dihedrals:
            self._condense_types(self.dihedral_types)
            self._condense_types(self.rb_torsion_types)
        self._condense_types(self.improper_periodic_types)
        self._condense_types(self.improper_types)
        self._condense_types(self.cmap_types)
        return self

    @staticmethod
    def _condense_types(typedict):
        """
        Loops through the given dict and condenses all types.

        Parameter
        ---------
        typedict : dict
            Type dictionary to condense
        """
        keylist = list(typedict.keys())
        for i in range(len(keylist) - 1):
            key1 = keylist[i]
            for j in range(i+1, len(keylist)):
                key2 = keylist[j]
                if typedict[key1] == typedict[key2]:
                    typedict[key2] = typedict[key1]

    @property
    def combining_rule(self):
        return self._combining_rule

    @combining_rule.setter
    def combining_rule(self, value):
        if value not in ('lorentz', 'geometric'):
            raise ValueError('combining_rule must be "lorentz" or "geometric"')
        self._combining_rule = value

    def typeify_templates(self):
        """ Assign atom types to atom names in templates """
        from parmed.modeller import ResidueTemplateContainer, ResidueTemplate
        for residue in itervalues(self.residues):
            if isinstance(residue, ResidueTemplateContainer):
                for res in residue:
                    for atom in res:
                        atom.atom_type = self.atom_types[atom.type]
            else:
                assert isinstance(residue, ResidueTemplate), 'Wrong type!'
                for atom in residue:
                    atom.atom_type = self.atom_types[atom.type]

def _find_ureybrad_key(urey):
    """
    Finds a key for a given Urey-Bradley by finding the middle atom in an angle.
    Raises a ParameterWarning if no middle atom found
    """
    a1, a2 = urey.atom1, urey.atom2
    shared_bond_partners = set(a1.bond_partners) & set(a2.bond_partners)
    if len({a.type for a in shared_bond_partners}) != 1:
        warnings.warn('Urey-Bradley %r shares multiple central atoms', ParameterWarning)
    return (a1.type, list(shared_bond_partners)[0].type, a2.type)

def _find_improper_keys(dih):
    """ Finds the central atom (i.e., that bonded to everything else) """
    assert dih.improper, 'Should not be called on non-improper!'
    all_atoms = {dih.atom1, dih.atom2, dih.atom3, dih.atom4}
    for atom in all_atoms:
        for oatom in all_atoms:
            if oatom is atom:
                continue
            if oatom not in atom.bond_partners:
                break
        else:
            # This *is* the central atom
            for key in permutations([a.type for a in all_atoms if a is not atom]):
                yield (key[0], key[1], atom.type, key[2])
            return # break out of the generator
    # If we got here, we found no central atom. *assume* it's the third spot already...
    for key in permutations([dih.atom1.type, dih.atom2.type, dih.atom4.type]):
        yield (key[0], key[1], dih.atom3.type, key[2])



"""
Pure python inversion of small matrices, to avoid requiring numpy or similar in SimTK.

This is part of the OpenMM molecular simulation toolkit originating from
Simbios, the NIH National Center for Physics-Based Simulation of
Biological Structures at Stanford, funded under the NIH Roadmap for
Medical Research, grant U54 GM072970. See https://simtk.org.

Portions copyright (c) 2012 Stanford University and the Authors.
Authors: Christopher M. Bruns
Contributors: Peter Eastman

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS, CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from __future__ import print_function, division, absolute_import
from parmed.utils.six.moves import range

import sys

def eye(size):
    """
    Returns identity matrix.

    >>> print(eye(3))
    [[1, 0, 0]
     [0, 1, 0]
     [0, 0, 1]]
    """
    result = []
    for row in range(size):
        r = []
        for col in range(size):
            if row == col:
                r.append(1)
            else:
                r.append(0)
        result.append(r)
    return MyMatrix(result)

def zeros(m, n=None):
    """
    Returns matrix of zeroes

    >>> print(zeros(3))
    [[0, 0, 0]
     [0, 0, 0]
     [0, 0, 0]]
    """
    if n is None:
        n = m
    result = []
    for row in range(m):
        r = []
        for col in range(n):
            r.append(0)
        result.append(r)
    return MyMatrix(result)

class MyVector(object):
    """
    Parent class of MyMatrix and type of Matrix Row.
    """
    def __init__(self, collection):
        if isinstance(collection, MyVector):
            self.data = collection.data
        else:
            self.data = collection

    def __str__(self):
        return str(self.data)

    def __repr__(self):
        return self.__class__.__name__ + "(" + repr(self.data) + ")"

    def __getitem__(self, key):
        return self.data[key]

    def __contains__(self, item):
        return item in self.data

    def __delitem__(self, key):
        del self.data[key]

    def __iter__(self):
        for item in self.data:
            yield item

    def __len__(self):
        return len(self.data)

    def __setitem__(self, key, value):
        self.data[key] = value

    def __rmul__(self, lhs):
        try:
            len(lhs)
            # left side is not scalar, delegate mul to that class
            return NotImplemented
        except TypeError:
            new_vec = []
            for element in self:
                new_vec.append(lhs * element)
            return self.__class__(new_vec)

class MyMatrix(MyVector):
    """
    Pure python linear algebra matrix for internal matrix inversion in UnitSystem.

    >>> m = MyMatrix([[1,0,],[0,1,]])
    >>> print(m)
    [[1, 0]
     [0, 1]]
    >>> print(~m)
    [[1.0, 0.0]
     [0.0, 1.0]]
    >>> print(eye(5))
    [[1, 0, 0, 0, 0]
     [0, 1, 0, 0, 0]
     [0, 0, 1, 0, 0]
     [0, 0, 0, 1, 0]
     [0, 0, 0, 0, 1]]
    >>> m = eye(5)
    >>> m[1][1]
    1
    >>> m[1:4]
    MyMatrixTranspose([[0, 0, 0],[1, 0, 0],[0, 1, 0],[0, 0, 1],[0, 0, 0]])
    >>> print(m[1:4])
    [[0, 0, 0]
     [1, 0, 0]
     [0, 1, 0]
     [0, 0, 1]
     [0, 0, 0]]
    >>> print(m[1:4][0:2])
    [[0, 1]
     [0, 0]
     [0, 0]]
    >>> m[1:4][0:2] = [[9,8],[7,6],[5,4]]
    >>> print(m)
    [[1, 0, 0, 0, 0]
     [9, 8, 0, 0, 0]
     [7, 6, 1, 0, 0]
     [5, 4, 0, 1, 0]
     [0, 0, 0, 0, 1]]
    """
    def numRows(self):
        return len(self.data)

    def numCols(self):
        if len(self.data) == 0:
            return 0
        else:
            return len(self.data[0])

    def __len__(self):
        return self.numRows()

    def __str__(self):
        result = ""
        start_char = "["
        for m in range(self.numRows()):
            result += start_char
            result += str(self[m])
            if m < self.numRows() - 1:
                result += "\n"
            start_char = " "
        result += "]"
        return result

    def __repr__(self):
        return 'MyMatrix(' + MyVector.__repr__(self) + ')'

    def is_square(self):
        return self.numRows() == self.numCols()

    def __iter__(self):
        for item in self.data:
            yield MyVector(item)

    def __getitem__(self, m):
        if isinstance(m, slice):
            return MyMatrixTranspose(self.data[m])
        else:
            return MyVector(self.data[m])

    def __setitem__(self, key, rhs):
        if isinstance(key, slice):
            self.data[key] = rhs
        else:
            assert len(rhs) == self.numCols()
            self.data[key] = MyVector(rhs)

    def __mul__(self, rhs):
        """
        Matrix multiplication.

        >>> a = MyMatrix([[1,2],[3,4]])
        >>> b = MyMatrix([[5,6],[7,8]])
        >>> print(a)
        [[1, 2]
         [3, 4]]
        >>> print(b)
        [[5, 6]
         [7, 8]]
        >>> print(a*b)
        [[19, 22]
         [43, 50]]

        """
        m = self.numRows()
        n = len(rhs[0])
        r = len(rhs)
        if self.numCols() != r:
            raise ArithmeticError("Matrix multplication size mismatch (%d vs %d)" % (self.numCols(), r))
        result = zeros(m, n)
        for i in range(m):
            for j in range(n):
                for k in range(r):
                    result[i][j] += self[i][k]*rhs[k][j]
        return result

    def __add__(self, rhs):
        """
        Matrix addition.

        >>> print(MyMatrix([[1, 2],[3, 4]]) + MyMatrix([[5, 6],[7, 8]]))
        [[6, 8]
         [10, 12]]
        """
        m = self.numRows()
        n = self.numCols()
        assert len(rhs) == m
        assert len(rhs[0]) == n
        result = zeros(m,n)
        for i in range(m):
            for j in range(n):
                result[i][j] = self[i][j] + rhs[i][j]
        return result

    def __sub__(self, rhs):
        """
        Matrix subtraction.

        >>> print(MyMatrix([[1, 2],[3, 4]]) - MyMatrix([[5, 6],[7, 8]]))
        [[-4, -4]
         [-4, -4]]
        """
        m = self.numRows()
        n = self.numCols()
        assert len(rhs) == m
        assert len(rhs[0]) == n
        result = zeros(m,n)
        for i in range(m):
            for j in range(n):
                result[i][j] = self[i][j] - rhs[i][j]
        return result

    def __pos__(self):
        return self

    def __neg__(self):
        m = self.numRows()
        n = self.numCols()
        result = zeros(m, n)
        for i in range(m):
            for j in range(n):
                result[i][j] = -self[i][j]
        return result

    def __invert__(self):
        """
        >>> m = MyMatrix([[1,1],[0,1]])
        >>> print(m)
        [[1, 1]
         [0, 1]]
        >>> print(~m)
        [[1.0, -1.0]
         [0.0, 1.0]]
        >>> print(m*~m)
        [[1.0, 0.0]
         [0.0, 1.0]]
        >>> print(~m*m)
        [[1.0, 0.0]
         [0.0, 1.0]]
        >>> m = MyMatrix([[1,0,0],[0,0,1],[0,-1,0]])
        >>> print(m)
        [[1, 0, 0]
         [0, 0, 1]
         [0, -1, 0]]
        >>> print(~m)
        [[1.0, 0.0, 0.0]
         [0.0, 0.0, -1.0]
         [0.0, 1.0, 0.0]]
        >>> print(m*~m)
        [[1.0, 0.0, 0.0]
         [0.0, 1.0, 0.0]
         [0.0, 0.0, 1.0]]
        >>> print(~m*m)
        [[1.0, 0.0, 0.0]
         [0.0, 1.0, 0.0]
         [0.0, 0.0, 1.0]]
        """
        assert self.is_square()
        if self.numRows() == 0:
            return self
        elif self.numRows() == 1:
            val = self[0][0]
            val = 1.0/val
            return MyMatrix([[val]])
        elif self.numRows() == 2: # 2x2 is analytic
            # http://en.wikipedia.org/wiki/Invertible_matrix#Inversion_of_2.C3.972_matrices
            a = self[0][0]
            b = self[0][1]
            c = self[1][0]
            d = self[1][1]
            determinant = a*d - b*c
            if determinant == 0:
                raise ArithmeticError("Cannot invert 2x2 matrix with zero determinant")
            else:
                return 1.0/(a*d - b*c) * MyMatrix([[d, -b],[-c, a]])
        else:
            # Gauss Jordan elimination from numerical recipes
            n = self.numRows()
            m1 = self.numCols()
            assert n == m1
            # Copy initial matrix into result matrix
            a = zeros(n, n)
            for i in range (0,n):
                for j in range (0,n):
                    a[i][j] = self[i][j]
            # These arrays are used for bookkeeping on the pivoting
            indxc = [0] * n
            indxr = [0] * n
            ipiv = [0] * n
            for i in range (0,n):
                big = 0.0
                for j in range (0,n):
                    if ipiv[j] != 1:
                        for k in range (0,n):
                            if ipiv[k] == 0:
                                if abs(a[j][k]) >= big:
                                    big = abs(a[j][k])
                                    irow = j
                                    icol = k
                ipiv[icol] += 1
                # We now have the pivot element, so we interchange rows...
                if irow != icol:
                    for l in range(n):
                        temp = a[irow][l]
                        a[irow][l] = a[icol][l]
                        a[icol][l] = temp
                indxr[i] = irow
                indxc[i] = icol
                if a[icol][icol] == 0:
                    raise ArithmeticError("Cannot invert singular matrix")
                pivinv = 1.0/a[icol][icol]
                a[icol][icol] = 1.0
                for l in range(n):
                    a[icol][l] *= pivinv
                for ll in range(n): # next we reduce the rows
                    if ll == icol:
                        continue # except the pivot one, of course
                    dum = a[ll][icol]
                    a[ll][icol] = 0.0
                    for l in range(n):
                        a[ll][l] -= a[icol][l]*dum
            # Unscramble the permuted columns
            for l in range(n-1, -1, -1):
                if indxr[l] == indxc[l]:
                    continue
                for k in range(n):
                    temp = a[k][indxr[l]]
                    a[k][indxr[l]] = a[k][indxc[l]]
                    a[k][indxc[l]] = temp
            return a

    def transpose(self):
        return MyMatrixTranspose(self.data)


class MyMatrixTranspose(MyMatrix):

    def transpose(self):
        return MyMatrix(self.data)

    def numRows(self):
        if len(self.data) == 0:
            return 0
        else:
            return len(self.data[0])

    def numCols(self):
        return len(self.data)

    def __getitem__(self, key):
        result = []
        for row in self.data:
            result.append(row[key])
        if isinstance(key, slice):
            return MyMatrix(result)
        else:
            return MyVector(result)

    def __setitem__(self, key, rhs):
        for n in range(len(self.data)):
            self.data[n][key] = rhs[n]

    def __str__(self):
        if len(self.data) == 0:
            return "[[]]"
        start_char = "["
        result = ""
        for m in range(len(self.data[0])):
            result += start_char
            result += "["
            sep_char = ""
            for n in range(len(self.data)):
                result += sep_char
                result += str(self.data[n][m])
                sep_char = ", "
            result += "]"
            if m < len(self.data[0]) - 1:
                result += "\n"
            start_char = " "
        result += "]"
        return result

    def __repr__(self):
        if len(self.data) == 0:
            return "MyMatrixTranspose([[]])"
        start_char = "["
        result = 'MyMatrixTranspose('
        for m in range(len(self.data[0])):
            result += start_char
            result += "["
            sep_char = ""
            for n in range(len(self.data)):
                result += sep_char
                result += repr(self.data[n][m])
                sep_char = ", "
            result += "]"
            if m < len(self.data[0]) - 1:
                result += ","
            start_char = ""
        result += '])'
        return result


# run module directly for testing
if __name__=='__main__':

    # Test the examples in the docstrings
    import doctest, sys
    doctest.testmod(sys.modules[__name__])



"""
This is a module that contains all of the necessary widgets that xParmEd will
use
"""
# Pull in all of Tkinter into our top-level namespace. This is common practice
# for Tkinter
from parmed.utils.six.moves.tkinter import *
from parmed.utils.six.moves.tkinter_messagebox import showerror, showinfo
from parmed.tools.argumentlist import ArgumentList
from parmed.tools.gui.guifiletools import file_chooser

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

class MessageWindow(Frame):
    """ 
    This is a window that just contains all of the messages printed out as 
    actions are completed.
    """
    def __init__(self, root):
        Frame.__init__(self, root)
        # Create the scrollable text area that cannot be modified
        self.text = Text(self, spacing3=5, width=60, wrap=WORD)
        scroller = Scrollbar(self, orient=VERTICAL)
        self.text.config(yscrollcommand=scroller.set)
        scroller.config(command=self.text.yview)
        self.text.config(state=DISABLED)
        self.text.grid(column=0, row=0, sticky=N+S+E+W)
        scroller.grid(column=1, row=0, sticky=N+S+E+W)
        # Make sure our window expands the way we want it to
        self.columnconfigure(0, weight=1)
        self.columnconfigure(1, weight=0)
        self.rowconfigure(0, weight=1)
        # Add the title
        self.write('Message log from ParmEd Actions\n')

    def add_line(self, text):
        self.write(text+'\n')

    def write(self, text):
        self.text.config(state=NORMAL)
        self.text.insert(END, text)
        self.text.config(state=DISABLED)

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

class BaseParmedWindow(Toplevel):
    """ 
    The base xParmEd window. It sits on top of a parent window and disables it
    until it is destroyed itself (in which case it re-enables it)
    """
    def __init__(self, amber_prmtop, title='Parmed Action Window'):
        Toplevel.__init__(self)
        self.amber_prmtop = amber_prmtop
        self.title(title)
        # Bind all events to this window
        self.grab_set()
   
#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

class RadioButtonWindow(BaseParmedWindow):
    """ 
    This is a window when the user can specify ONE option of a group of options
    with a single set of radio buttons
    """
    def __init__(self, amber_prmtop, title, desc, variable, button_namelist):
        BaseParmedWindow.__init__(self, amber_prmtop, title)
        self.variable = variable
        # Add a label describing what we're doing
        label = Label(self, text=desc, justify=LEFT, pady=5, padx=10)
        label.grid(column=0, row=0)
        frame = Frame(self, pady=5, padx=10)
        for i, name in enumerate(button_namelist):
            radio = Radiobutton(frame, variable=variable, text=name, value=name)
            radio.grid(column=0, row=i, sticky=W)
        frame.grid(column=0, row=1, sticky=W)
        # Give users a way out
        button = Button(self, text='OK', command=self.ok_destroy)
        button.grid(column=0, row=2, sticky=N+S+E+W)
        button = Button(self, text='Cancel', command=self.cancel_destroy)
        button.grid(column=0, row=3, sticky=N+S+E+W)

    def cancel_destroy(self):
        self.variable.set('')
        self.destroy()

    def ok_destroy(self):
        if not self.variable.get():
            showinfo('Missing Variable', "You did not choose an option.")
        self.destroy()

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

class MaskEntry(Frame):
    """
    This class is used to accept a Mask from a user. It gives them the
    option of inspecting their mask to see what atoms they will be selecting
    """
    def __init__(self, root, amber_prmtop, desc, var, restore_to):
        self.root = root
        self.restore_to = restore_to
        self.var = var
        self.amber_prmtop = amber_prmtop
        Frame.__init__(self, root)
        # Put a label for the entry
        label = Label(self, text=desc)
        label.grid(column=0, row=0, sticky=N+S+E+W)
        # Put an entry field in the frame
        self.entry = Entry(self, width=40, textvariable=var)
        self.entry.grid(column=0, row=1, sticky=N+S+E+W)
        # Put a button underneath to evaluate the mask
        self.button = Button(self, text='Evaluate Current Mask', 
                             command=self.evaluate_me)
        self.button.grid(column=0, row=2, sticky=N+S+E+W)
        # Now make a toplevel window for the evaluation, but keep it withdrawn
        # until it's needed
        self.window = Toplevel(root)
        self.text = ScrollText(self.window, restore_to, spacing3=5, padx=5,
                               pady=5, width=100, height=20)
        self.text.pack(fill=BOTH, expand=1)
        self.window.withdraw()
        self.window.resizable(True, True)
        self.window.transient(self)

    def evaluate_me(self):
        """ Evaluates the mask and dumps the contents into a Text box """
        from parmed.tools import printDetails
        from parmed.exceptions import MaskError
        mask = self.var.get().strip()
        if not mask: return
        # Clear out the text
        self.window.deiconify()
        self.text.clear()
        try: 
            maskstr = printDetails(self.amber_prmtop, ArgumentList(mask))
            self.text.write(str(maskstr))
        except MaskError:
            self.text.write('Bad mask string [%s]' % mask)
        self.window.grab_set()

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

class ActionWindow(BaseParmedWindow):
    """ A window with a bunch of user-entry fields
    
    Parameters
    ----------
    title : str
        Title that should be placed on the top bar of the window
    amber_prmtop : :class:`ParmList`
        List of topology files being modified
    widget_list : list of (str, str)
        Widget class names, descriptions of widgets
    var_list : list of XyzVar
        List of Var classes (StringVar, IntVar, FloatVar) corresponding to the
        entries in the widget_list
    description : str
        Description of the action being performed
    """
    def __init__(self, title, amber_prmtop, widget_list, var_list, description):
        self.cancelled = False
        self.var_list = var_list
        # widget_list: list of tuples having widget class names and description
        # title: title of the window (action name)
        # var_list: list of variables that go in each widget
        # description: Description to print out on the window
        BaseParmedWindow.__init__(self, amber_prmtop, title)
        # Add description label
        main_desc = Label(self, text=description, justify=CENTER,
                          pady=5, padx=10)
        main_desc.grid(column=0, row=0, sticky=N+S+E+W)
        # Make a frame to put all of our widgets in
        main_frame = Frame(self, padx=10, pady=5)
        # Now add all of our widgets to our 
        for i, wtemp in enumerate(widget_list):
            wname, wdesc = wtemp[0], wtemp[1]
            var = var_list[i]
            local_frame = Frame(main_frame)
            if wname == 'MaskEntry':
                mywidget = MaskEntry(local_frame, self.amber_prmtop,
                                     wdesc, var, self)
                mywidget.grid(row=0, column=0, sticky=N+S+E+W)
            elif wname == 'Entry':
                wlab = Label(local_frame, text=wdesc)
                wlab.grid(row=0, column=0, sticky=N+S+E+W)
                mywidget = Entry(local_frame, width=40, textvariable=var)
                mywidget.grid(row=1, column=0, sticky=N+S+E+W)
            elif wname == 'Spinbox':
                wlab = Label(local_frame, text=wdesc)
                wlab.grid(row=0, column=0, sticky=N+S+E+W)
                mywidget = Spinbox(local_frame, textvariable=var,
                                   values=wtemp[2:])
                mywidget.grid(row=1, column=0, sticky=N+S+E+W)
            elif wname == 'Checkbutton':
                mywidget = Checkbutton(local_frame, text=wdesc, variable=var,
                                    onvalue='yes', offvalue='no')
                mywidget.grid(row=0, column=0, sticky=N+S+E+W)
            elif wname == 'FileSelector':
                def callback():
                    file_chooser('C4 Parameter File',
                            extensions=[('All files', '*')], set_var=var)
                mywidget = Button(local_frame, textvariable=var,
                                  command=callback)
                wlab = Label(local_frame, text=wdesc)
                wlab.grid(row=0, column=0, sticky=N+S+E+W)
                mywidget.grid(row=1, column=0, sticky=N+S+E+W)
            else:
                showerror('Error!', '%s not implemented yet!' % wname)
                self.destroy()
            local_frame.grid(row=i//2, column=i%2, sticky=N+S+E+W)

        # Now pack the main frame
        main_frame.grid(column=0, row=1, sticky=N+S+E+W)
        # Now make the OK and Cancel buttons
        ok_can_frame = Frame(self, padx=10, pady=5)
        ok_can_frame.grid(column=0, row=2)
        button = Button(ok_can_frame, text='OK', command=self.destroy)
        button.grid(row=0, column=0, sticky=N+S+E+W, padx=10)
        button = Button(ok_can_frame, text='Cancel', command=self.cancel)
        button.grid(row=0, column=1, sticky=N+S+E+W, padx=10)

    def cancel(self):
        """ Clears out the variables and bails out """
        self.cancelled = True
        for var in self.var_list: var.set('')
        self.destroy()

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

class ScrollText(Frame):
    """ A scrollable text (in the vertical direction) """
    ok_button_label = 'OK / Go Back'
    def __init__(self, root, restore_to, **options):
        self.restore_to = restore_to
        Frame.__init__(self, root)
        self.text = Text(self, **options)
        self.text.grid(column=0, row=0, sticky=N+E+S+W)
        self.scroller = Scrollbar(self, orient=VERTICAL,
                                  command=self.text.yview)
        self.scroller.grid(column=1, row=0, sticky=N+S+E+W)
        self.text.config(yscrollcommand=self.scroller.set)
        self.button = Button(self, text=self.ok_button_label, command=self.kill)
        self.button.grid(column=0, row=1, sticky=E+W, columnspan=2)
        root.protocol('WM_DELETE_WINDOW', self.kill)
        self.columnconfigure(0, weight=1)
        self.columnconfigure(1, weight=0)
        self.rowconfigure(0, weight=1)
        self.text.config(state=DISABLED)

    def write(self, text):
        """ Writes text to the text window """
        self.text.config(state=NORMAL)
        self.text.insert(END, str(text))
        self.text.config(state=DISABLED)

    def clear(self):
        """ Clears all text from the window """
        self.text.config(state=NORMAL)
        self.text.delete(1.0, END)
        self.text.config(state=DISABLED)

    def kill(self):
        self.master.withdraw()
        # Have the old window grab the focus
        if self.restore_to: 
            self.restore_to.grab_set()
        else: 
            self.grab_release()

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

class ExitingScrollText(ScrollText):
    """ A ScrollText that will quit after its OK button is clicked """
    ok_button_label = 'OK / Quit'
    def kill(self):
        self.master.destroy()



"""
This module handles adding coarse graining data to topology files set up by
Luciana Capece. It has been implemented here in parmed by Jason Swails, but all
ideas were Luciana's. Here is a description of what this section does to
topology files:

ANGLE_FORCE_CONSTANT has been rigged to provide an integer index to the specific
parameter found in the given parameter file. This will be used to set up 4
different (new) prmtop sections: ANGLE_COEF_A, ANGLE_COEF_B, ANGLE_COEF_C, and
ANGLE_COEF_D. ANGLE_FORCE_CONSTANT now becomes useless, so I will add a comment
to the topology file saying such.

DIHEDRAL_FORCE_CONSTANT has been rigged in much the same way, creating 8 new
sections: DIHEDRAL_AMPLITUDE_1,2,3,4 and DIHEDRAL_PHASE_1,2,3,4.
DIHEDRAL_FORCE_CONSTANT, DIHEDRAL_PHASE, and DIHEDRAL_PERIODICITY become useless
and are commented as such (though not removed).

This module sets up ANGLE and DIHEDRAL classes to make printing them easier,
then parses the parameter file, then sets the respective sections of the prmtop
file.
"""
from parmed.utils.six.moves import range
import warnings

#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

class Angle(object):
    """ Angle between 3 bonded atoms """
    def __init__(self, atom1, atom2, atom3, acoef, bcoef, ccoef, dcoef):
        self.atom1 = atom1.strip()
        self.atom2 = atom2.strip()
        self.atom3 = atom3.strip()
        self.acoef = float(acoef)
        self.bcoef = float(bcoef)
        self.ccoef = float(ccoef)
        self.dcoef = float(dcoef)

    def __eq__(self, other):
        """
        Defines if 2 angles are equal: the middle atom has to be the same, and
        the 2 outer ones have to be the same in either order, or they can be
        wildcards (X)
        """
        # First see if we are looking to match the angles in forward or reverse
        reversed_angles = False
        if self.atom2 != other.atom2: return False
        if self.atom1 != other.atom1:
            if self.atom1 != other.atom3:
                if self.atom1 != 'X' and other.atom1 != 'X':
                    if other.atom3 == 'X':
                        reversed_angles = True
                    else:
                        return False
                # end if self.atom1 != 'X' and other.atom1 != 'X'
            else:
                reversed_angles = True
        # Now check the rest (non-reversed case first)
        if not reversed_angles:
            if (self.atom3 != other.atom3 and self.atom3 != 'X' and
                other.atom3 != 'X'):
                return False
        else: # reversed angles
            if (self.atom3 != other.atom1 and self.atom3 != 'X' and
                other.atom1 != 'X'):
                return False
        # If we passed all these tests so far, then we must be the same!
        return True

#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

class Dihedral(object):
    """ Dihedral between 4 bonded atoms """
    def __init__(self, atom1, atom2, atom3, atom4, 
                 ampl1, ampl2, ampl3, ampl4,
                 phase1, phase2, phase3, phase4):
        self.atom1, self.atom2 = atom1, atom2
        self.atom3, self.atom4 = atom3, atom4
        self.ampl1, self.ampl2 = ampl1, ampl2
        self.ampl3, self.ampl4 = ampl3, ampl4
        self.phase1, self.phase2 = phase1, phase2
        self.phase3, self.phase4 = phase3, phase4

    def __eq__(self, other):
        """
        2 Dihedrals are equal if they go in forward/reverse order and all atom
        types either match or we have wild cards (which can only be on the ends)
        """
        reversed_angles = False
        if self.atom2 != other.atom2 or self.atom3 != other.atom3:
            if self.atom2 != other.atom3 or self.atom3 != other.atom2:
                return False
            else:
                reversed_angles = True
        if not reversed_angles:
            if self.atom1 != other.atom1:
                if self.atom1 != 'X' and other.atom1 != 'X':
                    return False
            if self.atom4 != other.atom4:
                if self.atom4 != 'X' and other.atom4 != 'X':
                    return False
        else:
            if self.atom1 != other.atom4:
                if self.atom1 != 'X' and other.atom4 != 'X':
                    return False
            if self.atom4 != other.atom1:
                if self.atom4 != 'X' and other.atom4 != 'X':
                    return False
        # There must be no problems if we made it this far
        return True

#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

def addCoarseGrain(parm, param_file):
    from parmed.tools.exceptions import CoarseGrainError
    """ Adds coarse graining sections to the prmtop """
    # Angles and Dihedrals dictionaries for storing parameters parsed from
    # parameter file
    angle_params = {}
    dihedral_params = {}
    # First parse the params file, filling dictionaries of angles and dihedrals
    # whose keys are their integer index values
    params = open(param_file, 'r')
    reading_angles = False
    reading_dihedrals = False
    file_line = params.readline()
    while file_line:
        # Filter out comments
        if '#' in file_line:
            file_line = file_line[:file_line.index('#')]
        # Filter out blank lines
        file_line = file_line.strip()
        if len(file_line) == 0: 
            file_line = params.readline()
            continue
        # See if we start reading ANGLEs yet:
        if file_line[:5].upper() == 'ANGLE':
            reading_angles = True
            reading_dihedrals = False
            file_line = params.readline()
            continue
        # See if we start reading DIHEDRALs yet
        if file_line[:8].upper() == 'DIHEDRAL':
            reading_dihedrals = True
            reading_angles = False
            file_line = params.readline()
            continue
        # Pass over any INDEX header lines -- they're just there for readability
        if (reading_dihedrals or reading_angles) and file_line[:5] == 'INDEX':
            file_line = params.readline()
            continue
        if reading_angles:
            line_parts = file_line.split()
            try:
                idx = int(line_parts[0])
                acoef = float(line_parts[1])
                bcoef = float(line_parts[2])
                ccoef = float(line_parts[3])
                dcoef = float(line_parts[4])
                atom_types = line_parts[5].split('-')
                angle_params[idx] = Angle(atom_types[0], atom_types[1],
                                          atom_types[2], acoef, bcoef, ccoef,
                                          dcoef)
            except ValueError as err:
                raise CoarseGrainError(
                        'Unexpected format in Coarse Grain angles. Expected '
                        'different data type: %s. See format specification'
                        % err)
            except IndexError:
                raise CoarseGrainError(
                        'Unexpected format in Coarse Grain parameter file. '
                        'Expected more data fields on the line. See format '
                        'specification')
            file_line = params.readline()
            continue
        if reading_dihedrals:
            line_parts = file_line.split()
            try:
                idx = int(line_parts[0])
                ampl1 = float(line_parts[1])
                ampl2 = float(line_parts[2])
                ampl3 = float(line_parts[3])
                ampl4 = float(line_parts[4])
                atom_types = line_parts[5].split('-')
                file_line = params.readline().strip()
                line_parts = file_line.split()
                phase1 = float(line_parts[0])
                phase2 = float(line_parts[1])
                phase3 = float(line_parts[2])
                phase4 = float(line_parts[3])
                dihedral_params[idx] = Dihedral(atom_types[0], atom_types[1],
                                                atom_types[2], atom_types[3],
                                                ampl1, ampl2, ampl3, ampl4,
                                                phase1, phase2, phase3, phase4)
            except ValueError as err:
                raise CoarseGrainError(
                        'Unexpected format in Coarse Grain dihedrals. Expected '
                        'different data type: %s. See format specification'
                        % err)
            except IndexError:
                raise CoarseGrainError(
                        'Unexpected format in Coarse Grain parameter file. '
                        'Expected more data fields on the line. See format '
                        'specification')
            file_line = params.readline()
            continue
        warnings.warn('Line (%s) ignored in Coarse Grain parameter file' %
                      file_line)
        file_line = params.readline()
    # End reading while

    # Now let's add comments to our topology file!
    parm.parm_comments['ANGLE_FORCE_CONSTANT'].append(
                'This section is only an index and is not used for Coarse '
                'grained topologies')
    parm.parm_comments['DIHEDRAL_FORCE_CONSTANT'].append(
                'This section is only an index and is not used for Coarse '
                'grained topologies')
    parm.parm_comments['DIHEDRAL_PERIODICITY'].append(
                'This section is not used for Coarse grained topologies')
    parm.parm_comments['DIHEDRAL_PHASE'].append(
                'This section is not used for Coarse grained topologies')
   
    # Now let's add our new sections
    parm.add_flag('ANGLE_COEF_A','5E16.8',parm.ptr('numang'),
                 comments='A Coefficient for Coarse grained force field')
    parm.add_flag('ANGLE_COEF_B','5E16.8',parm.ptr('numang'),
                 comments='B Coefficient for Coarse grained force field')
    parm.add_flag('ANGLE_COEF_C','5E16.8',parm.ptr('numang'),
                 comments='C Coefficient for Coarse grained force field')
    parm.add_flag('ANGLE_COEF_D','5E16.8',parm.ptr('numang'),
                 comments='D Coefficient for Coarse grained force field')
    parm.add_flag('DIHEDRAL_AMPLITUDE_1','5E16.8',parm.ptr('nptra'),
                 comments='1st Dihedral Amplitude for coarse grained force '
                          'field')
    parm.add_flag('DIHEDRAL_AMPLITUDE_2','5E16.8',parm.ptr('nptra'),
                 comments='2nd Dihedral Amplitude for coarse grained force '
                          'field'
    )
    parm.add_flag('DIHEDRAL_AMPLITUDE_3','5E16.8',parm.ptr('nptra'),
                 comments='3rd Dihedral Amplitude for coarse grained force '
                          'field'
    )
    parm.add_flag('DIHEDRAL_AMPLITUDE_4','5E16.8',parm.ptr('nptra'),
                 comments='4th Dihedral Amplitude for coarse grained force '
                          'field'
    )
    parm.add_flag('DIHEDRAL_PHASE_1','5E16.8',parm.ptr('nptra'),
                 comments='1st Dihedral Phase for coarse grained force field')
    parm.add_flag('DIHEDRAL_PHASE_2','5E16.8',parm.ptr('nptra'),
                 comments='2nd Dihedral Phase for coarse grained force field')
    parm.add_flag('DIHEDRAL_PHASE_3','5E16.8',parm.ptr('nptra'),
                 comments='3rd Dihedral Phase for coarse grained force field')
    parm.add_flag('DIHEDRAL_PHASE_4','5E16.8',parm.ptr('nptra'),
                 comments='4th Dihedral Phase for coarse grained force field')

    for i in range(len(parm.parm_data['ANGLE_FORCE_CONSTANT'])):
        try: 
            index = int(parm.parm_data['ANGLE_FORCE_CONSTANT'][i])
            angl = angle_params[index]
        except KeyError:
            raise CoarseGrainError('Missing angle parameters for index %d' %
                                   int(index))
        parm.parm_data['ANGLE_COEF_A'][i] = angl.acoef
        parm.parm_data['ANGLE_COEF_B'][i] = angl.bcoef
        parm.parm_data['ANGLE_COEF_C'][i] = angl.ccoef
        parm.parm_data['ANGLE_COEF_D'][i] = angl.dcoef
   
    for i in range(len(parm.parm_data['DIHEDRAL_FORCE_CONSTANT'])):
        try:
            index = int(parm.parm_data['DIHEDRAL_FORCE_CONSTANT'][i])
            dihe = dihedral_params[index]
        except KeyError:
            raise CoarseGrainError('Missing dihedral parameters for index %d' %
                                   int(index))
        parm.parm_data['DIHEDRAL_AMPLITUDE_1'][i] = dihe.ampl1
        parm.parm_data['DIHEDRAL_AMPLITUDE_2'][i] = dihe.ampl2
        parm.parm_data['DIHEDRAL_AMPLITUDE_3'][i] = dihe.ampl3
        parm.parm_data['DIHEDRAL_AMPLITUDE_4'][i] = dihe.ampl4
        parm.parm_data['DIHEDRAL_PHASE_1'][i] = dihe.phase1
        parm.parm_data['DIHEDRAL_PHASE_2'][i] = dihe.phase2
        parm.parm_data['DIHEDRAL_PHASE_3'][i] = dihe.phase3
        parm.parm_data['DIHEDRAL_PHASE_4'][i] = dihe.phase4



"""
A package dealing with RDKit integration.
"""
from parmed.rdkit.rdkit import RDKit

__all__ = ['load_rdkit', 'from_smiles', 'from_sdf']

load_rdkit = RDKit.load
from_smiles = RDKit.from_smiles
from_sdf = RDKit.from_sdf



"""
This module contains the necessary machinery to add the LENNARD_JONES_CCOEF to a
topology file given a list of atomic polarizabilities
"""
from __future__ import division, print_function

from parmed.utils.six.moves import range, zip
from parmed.tools.exceptions import LJ12_6_4Error, DuplicateParamWarning
import warnings

WATER_POL = 1.444 # Polarizability of water

DEFAULT_C4_PARAMS = {
        'TIP3P' : {'Li1': 27.0, 'Na1': 0.0, 'K1': 8.0, 'Rb1': 0.0, 'Cs1': 2.0,
                   'Tl1': 50.0, 'Cu1': 7.0, 'Ag1': 83.0, 'F-1': -27.0,
                   'Cl-1': -38.0, 'Br-1': -39.0, 'I-1': -45.0, 'Be2': 186.5,
                   'Cu2': 290.9, 'Ni2': 212.8, 'Zn2': 231.6, 'Co2': 209.7,
                   'Cr2': 136.8, 'Fe2': 163.0, 'Mg2': 132.9, 'V2': 195.7,
                   'Mn2': 146.1, 'Hg2': 288.8, 'Cd2': 185.6, 'Ca2': 87.3,
                   'Sn2': 187.9, 'Sr2': 82.7, 'Ba2': 71.9, 'Al3': 399.0,
                   'Fe3': 428.0, 'Cr3': 258.0, 'In3': 347.0, 'Tl3': 456.0,
                   'Y3': 216.0, 'La3': 152.0, 'Ce3': 230.0, 'Pr3': 264.0,
                   'Nd3': 213.0, 'Sm3': 230.0, 'Eu3': 259.0, 'Gd3': 198.0,
                   'Tb3': 235.0, 'Dy3': 207.0, 'Er3': 251.0, 'Tm3': 282.0,
                   'Lu3': 249.0, 'Hf4': 827.0, 'Zr4': 761.0, 'Ce4': 706.0,
                   'U4': 1034.0, 'Pu4': 828.0, 'Th4': 512.0},
        'TIP4PEW' : {'Li1': 36.0, 'Na1': 9.0, 'K1': 24.0, 'Rb1': 13.0,
                     'Cs1': 16.0, 'Tl1': 65.0, 'Cu1': 21.0, 'Ag1': 94.0,
                     'F-1': -67.0, 'Cl-1': -66.0, 'Br-1': -68.0, 'I-1': -62.0,
                     'Be2': 228.5, 'Cu2': 339.2, 'Ni2': 259.2, 'Zn2': 272.3,
                     'Co2': 252.8, 'Cr2': 177.4, 'Fe2': 201.1, 'Mg2': 180.5,
                     'V2': 244.8, 'Mn2': 192.3, 'Hg2': 335.2, 'Cd2': 233.7,
                     'Ca2': 128.0, 'Sn2': 231.4, 'Sr2': 118.9, 'Ba2': 112.5,
                     'Al3': 488.0, 'Fe3': 519.0, 'Cr3': 322.0, 'In3': 425.0,
                     'Tl3': 535.0, 'Y3': 294.0, 'La3': 243.0, 'Ce3': 315.0,
                     'Pr3': 348.0, 'Nd3': 297.0, 'Sm3': 314.0, 'Eu3': 345.0,
                     'Gd3': 280.0, 'Tb3': 313.0, 'Dy3': 298.0, 'Er3': 328.0,
                     'Tm3': 356.0, 'Lu3': 331.0, 'Hf4': 956.0, 'Zr4': 895.0,
                     'Ce4': 835.0, 'U4': 1183.0, 'Pu4': 972.0, 'Th4': 625.0},
        'SPCE' :  {'Li1': 33.0, 'Na1': 6.0, 'K1': 19.0, 'Rb1': 7.0, 'Cs1': 12.0,
                   'Tl': 61.0, 'Cu1': 9.0, 'Ag1': 92.0, 'F-1': -53.0,
                   'Cl-1': -55.0, 'Br-1': -51.0, 'I-1': -51.0, 'Be2': 188.1,
                   'Cu2': 304.4, 'Ni2': 205.2, 'Zn2': 231.2, 'Co2': 209.2,
                   'Cr2': 131.2, 'Fe2': 155.4, 'Mg2': 122.2, 'V2': 206.6,
                   'Mn2': 154.9, 'Hg2': 300.2, 'Cd2': 198.8, 'Ca2': 89.0,
                   'Sn2': 201.1, 'Sr2': 96.3, 'Ba2': 85.8, 'Al3': 406.0,
                   'Fe3': 442.0, 'Cr3': 254.0, 'In3': 349.0, 'Tl3': 455.0,
                   'Y3': 209.0, 'La3': 165.0, 'Ce3': 242.0, 'Pr3': 272.0,
                   'Nd3': 235.0, 'Sm3': 224.0, 'Eu3': 273.0, 'Gd3': 186.0,
                   'Tb3': 227.0, 'Dy3': 206.0, 'Er3': 247.0, 'Tm3': 262.0,
                   'Lu3': 247.0, 'Hf4': 810.0, 'Zr4': 760.0, 'Ce4': 694.0,
                   'U4': 1043.0, 'Pu4': 828.0, 'Th4': 513.0}
}

def params1264(parm, mask, c4file, watermodel, polfile, tunfactor):
   
    from parmed import periodic_table as pt

    global DEFAULT_C4_PARAMS, WATER_POL

    try:
        pollist = _get_params(polfile)
    except ValueError:
        raise LJ12_6_4Error('Bad polarizability file %s. Expected a file with '
                            '2 columns: <Amber Atom Type> <Polarizability>' %
                            polfile)
   
    if c4file is None:
        c4list = DEFAULT_C4_PARAMS[watermodel]
    else:
        try:
            c4list = _get_params(c4file)
        except ValueError:
            raise LJ12_6_4Error('Bad C4 parameter file %s. Expected a file '
                                'with 2 columns: <Atom Element> '
                                '<C4 Parameter>' % c4file)


    print("***********************************************************")
    # Determine which atom type was treated as the center metal ion
    mettypdict = dict()
    for i in mask.Selected():
        mettypind = parm.parm_data['ATOM_TYPE_INDEX'][i]
        metchg = parm.parm_data['CHARGE'][i]
        if mettypind in mettypdict: continue
        mettypdict[mettypind] = (parm.atoms[i].atomic_number, int(metchg))
        print("The selected metal ion is %s" %
              pt.Element[parm.atoms[i].atomic_number])
    mettypinds = sorted(mettypdict.keys())

    # 1. Get the dict between AMBER_ATOM_TYPE and ATOM_TYPE_INDEX for all 
    # the atoms in prmtop file

    ntypes = parm.pointers['NTYPES']
    typs = parm.parm_data['AMBER_ATOM_TYPE']
    typinds = parm.parm_data['ATOM_TYPE_INDEX']

    typdict = {}
    for ty, ind in zip(typs, typinds):
        if ind not in typdict:
            typdict[ind] = [ty]
        elif ty in typdict[ind]:
            continue
        else:
            typdict[ind].append(ty)

    for i in range(1, ntypes+1):
        if i not in typinds:
            typdict[i] = []

    # 2.Generate the C4 term for each atom type pair
    result = [0.0 for i in parm.parm_data['LENNARD_JONES_ACOEF']]

    for mettypind in mettypinds:
        # Obtain the C4 parameters
        c4 = c4list[pt.Element[mettypdict[mettypind][0]] +
                    str(mettypdict[mettypind][1])]
        i = mettypind - 1
        for j in range(1, ntypes+1):
            jj = j - 1
            attypjs = typdict[j]
            if len(attypjs) >= 1:
                # Get polarizability
                for k, typjs in enumerate(attypjs):
                    if k == 0:
                        try:
                            pol = pollist[typjs]
                        except KeyError:
                            raise LJ12_6_4Error("Could not find parameters for "
                                                "ATOM_TYPE %s" % typjs )
                    else:
                        try:
                            anthpol = pollist[typjs]
                            if anthpol != pol:
                                raise LJ12_6_4Error(
                                        'Polarizability parameter of '
                                        'AMBER_ATOM_TYPE %s is not the same '
                                        'as that of AMBER_ATOM_TYPE %s, but '
                                        'their VDW parameters are the same. ' %
                                        (attypjs[0], typjs)
                                )
                        except KeyError:
                            raise LJ12_6_4Error("Could not find parameters for "
                                                "ATOM_TYPE %s" % typjs)
                # Get index
                if jj < i:
                    idx = parm.parm_data['NONBONDED_PARM_INDEX'][ntypes*jj+i]-1
                else:
                    idx = parm.parm_data['NONBONDED_PARM_INDEX'][ntypes*i+jj]-1

                # Caculate C4 terms
                if attypjs == ['OW']:
                    # There is only one C4 term exist between water and a
                    # certain ion
                    result[idx] = c4
                else:
                    # There are two C4 terms need to add together between two
                    # different ions
                    result[idx] += c4 / WATER_POL * pol * tunfactor
    return result

def _get_params(fname):
    params = dict()

    with open(fname, 'r') as f:
        for line in f:
            atomtype, param = line.split()[:2]
            param = float(param)
            if atomtype in params and abs(param - params[atomtype]) > 0.0001:
                warnings.warn('Atom type %s has multiple parameters in %s.' %
                              (atomtype, fname), DuplicateParamWarning)
            params[atomtype] = param

        return params
